#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import sasoptpy
from sasoptpy.core import Expression


class Constraint(Expression):
    """
    Creates a linear or quadratic constraint for optimization models

    Constraints should be created by adding logical relations to
    :class:`Expression` objects.

    Parameters
    ----------
    exp : :class:`Expression`
        A logical expression that forms the constraint
    direction : string, optional
        Direction of the logical expression

        Possible values are

        - `E` for equality (=) constraints
        - `L` for less than or euqal to (<=) constraints
        - `G` for greater than or equal to (>=) constraints

    name : string, optional
        Name of the constraint object
    crange : float, optional
        Range for ranged constraints

    Examples
    --------

    >>> x = so.Variable(name='x')
    >>> y = so.Variable(name='y')
    >>> c1 = so.Constraint( 3 * x - 5 * y <= 10, name='c1')
    >>> print(repr(c1))
    sasoptpy.Constraint( -  5.0 * y  +  3.0 * x  <=  10, name='c1')

    >>> c2 = so.Constraint( - x + 2 * y - 5, direction='L', name='c2')
    sasoptpy.Constraint( -  x  +  2.0 * y  <=  5, name='c2')

    Notes
    -----

    * A constraint can be generated in two different ways:

      - Using the :func:`sasoptpy.Model.add_constraint` method

         >>> m = so.Model(name='m')
         >>> c1 = m.add_constraint(3 * x - 5 * y <= 10, name='c1')
         >>> print(repr(c1))
         sasoptpy.Constraint( -  5.0 * y  +  3.0 * x  <=  10, name='c1')

      - Using the constructor

         >>> c1 = sasoptpy.Constraint(3 * x - 5 * y <= 10, name='c1')
         >>> print(repr(c1))
         sasoptpy.Constraint( -  5.0 * y  +  3.0 * x  <=  10, name='c1')

    * The same constraint can be included into other models using the
      :func:`Model.include` method.

    See also
    --------
    :func:`sasoptpy.Model.add_constraint`
    """

    @sasoptpy.class_containable
    def __init__(self, exp, direction=None, name=None, crange=None, condition=None):
        super().__init__(name=name)
        if name is not None:
            self._objorder = sasoptpy.util.get_creation_id()

        if exp._name is None:
            self._linCoef = exp.get_member_dict()
        else:
            self._copy_coef(exp)

        if direction is None:
            self._direction = exp._direction
        else:
            self._direction = direction

        if crange is not None:
            self._range = crange
        else:
            self._range = exp._range

        self.add_conditions(condition)

        if sasoptpy.conditions:
            self.add_conditions(sasoptpy.conditions)

        self._key = None
        self._parent = None
        self._block = None
        self._temp = False
        self._shadow = False

    def get_parent_reference(self):
        return self._parent, self._key

    def update_var_coef(self, var, value):
        """
        Updates the coefficient of a variable inside the constraint

        Parameters
        ----------
        var : :class:`Variable`
            Variable to be updated
        value : float
            Coefficient of the variable in the constraint

        Examples
        --------

        >>> c1 = so.Constraint(exp=3 * x - 5 * y <= 10, name='c1')
        >>> print(repr(c1))
        sasoptpy.Constraint( -  5.0 * y  +  3.0 * x  <=  10, name='c1')
        >>> c1.update_var_coef(x, -1)
        >>> print(repr(c1))
        sasoptpy.Constraint( -  5.0 * y  -  x  <=  10, name='c1')

        See also
        --------
        :func:`sasoptpy.Model.set_coef`

        """
        varname = var.get_name()
        if varname in self._linCoef:
            self._linCoef[varname]['val'] = value
        else:
            self._linCoef[varname] = {'ref': var, 'val': value}

    def set_rhs(self, value):
        """
        Changes the constant value (right-hand side) of a constraint

        Parameters
        ----------
        value : float
            New right-hand side value for the constraint

        Examples
        --------

        >>> x = m.add_variable(name='x')
        >>> y = m.add_variable(name='y')
        >>> c = m.add_constraint(x + 3*y <= 10, name='con_1')
        >>> print(c)
        x  +  3.0 * y  <=  10
        >>> c.set_rhs(5)
        >>> print(c)
        x  +  3.0 * y  <=  5

        """
        self._linCoef['CONST']['val'] = -value

    def set_direction(self, direction):
        """
        Changes the direction of a constraint

        Parameters
        ----------
        direction : string
            Direction of the constraint

            Possible values are

            - `E` for equality (=) constraints
            - `L` for less than or euqal to (<=) constraints
            - `G` for greater than or equal to (>=) constraints

        Examples
        --------

        >>> c1 = so.Constraint(exp=3 * x - 5 * y <= 10, name='c1')
        >>> print(repr(c1))
        sasoptpy.Constraint( 3.0 * x  -  5.0 * y  <=  10, name='c1')
        >>> c1.set_direction('G')
        >>> print(repr(c1))
        sasoptpy.Constraint( 3.0 * x  -  5.0 * y  >=  10, name='c1')

        """
        if direction in ['E', 'L', 'G']:
            self._direction = direction
        else:
            raise ValueError(
                'Not a valid constraint direction {}'.format(direction))

    def set_block(self, block_number):
        """
        Sets the decomposition block number for a constraint

        Parameters
        ----------
        block_number : int
            Block number of the constraint

        Examples
        --------

        >>> c1 = m.add_constraints((x + 2 * y[i] <= 5 for i in NODES),
                                   name='c1')
        >>> for i in NODES:
              c1[i].set_block(i)

        """
        self._block = block_number

    def get_value(self, rhs=False):
        """
        Returns the current value of the constraint

        Parameters
        ----------
        rhs : boolean, optional
            When set to `True`, includes constant values to the value of the
            constraint. Default is `False`.

        Examples
        --------

        >>> x = so.Variable(name='x', init=2)
        >>> c = so.Constraint(x ** 2 + 2 * x <= 15, name='c')
        >>> print(c.get_value())
        8
        >>> print(c.get_value(rhs=True))
        -7

        """
        v = super().get_value()
        if rhs:
            return v
        else:
            v -= self._linCoef['CONST']['val']
            return v

    def get_dual(self):
        """
        Returns the dual value if exists

        Returns
        -------
        dual : float
            Dual value of the constraint

        """
        return self._dual

    def add_conditions(self, conditions):
        if conditions is not None:
            for i in conditions:
                self.sym.add_condition(i)

    def _set_info(self, parent, key, shadow=False):
        self._parent = parent
        self._key = key
        self._shadow = shadow

    def __and__(self, other):
        r = sasoptpy.abstract.Condition(left=self, c_type='and', right=other)
        return r

    def __or__(self, other):
        r = sasoptpy.abstract.Condition(left=self, c_type='or', right=other)
        return r

    def __bool__(self):
        if sasoptpy.conditions is not None:
           sasoptpy.conditions.append(self)
        return True

    def __eq__(self, other):
        return id(self) == id(other)

    def _cond_expr(self):
        return self._get_constraint_expr()

    def _get_optmodel_name(self):
        if self._parent is not None:
            return self.get_parent_reference()[0].get_name() + \
                   sasoptpy.util.package_utils._to_optmodel_loop(
                           self._key, self)
        else:
            return self._name

    def _get_definition_tag(self):
        if self._parent is None:
            return 'con {} : '.format(self.get_name())
        return ''

    def get_range_expr(self):
        if self._range != 0:
            return '{} <= '.format(
                sasoptpy.util.get_in_digit_format(
                    -self.get_member_value('CONST')
                )
            )
        return ''

    def get_right_sign(self):
        if self._range != 0 and self._direction == 'E':
            return ' {} '.format(sasoptpy.util.get_direction_str('L'))
        else:
            return ' {} '.format(
            sasoptpy.util.get_direction_str(self._direction))

    def _expr(self):
        return self._get_constraint_expr()

    def _get_constraint_expr(self):
        left_expr = self.get_range_expr()
        expr = super()._expr()
        right_sign = self.get_right_sign()
        constant_side = '{}'.format(
            sasoptpy.util.get_in_digit_format(
                -self.get_member_value('CONST') + self._range))
        return left_expr + expr + right_sign + constant_side

    def _get_name_list(self):
        from sasoptpy.util.package_utils import \
            get_subindices, get_subindices_optmodel_format, \
            get_iterators, get_iterators_optmodel_format, _insert_brackets

        if self._key:
            group_types = self._parent.get_group_types()
            abstract_keys = [i[1] for i in enumerate(self._key)
                             if group_types[i[0]] == sasoptpy.ABSTRACT]
            concrete_keys = [i[1] for i in enumerate(self._key)
                             if group_types[i[0]] == sasoptpy.CONCRETE]

            parent_name = self._parent.get_name()
            iterators_exp = get_iterators_optmodel_format(abstract_keys)
            subindex_exp = get_subindices_optmodel_format(concrete_keys)
            con_name = _insert_brackets(parent_name + subindex_exp, tuple(abstract_keys))
            if iterators_exp == '':
                return [con_name]
            else:
                return [iterators_exp + ' ' + con_name]
        else:
            return [self._name]

    def _defn(self):
        s = self._get_definition_tag()
        s += self._get_constraint_expr()
        if self._parent is None:
            s += ';'
        return s

    def __str__(self):
        """
        Generates a representation string
        """
        s = super().__str__()
        if self._direction == 'E':
            s += ' == '
        elif self._direction == 'L':
            s += ' <= '
        elif self._direction == 'G':
            s += ' >= '
        elif self._direction == 'NE':
            s += ' ne '
        else:
            try:
                s += ' ' + sasoptpy._relation_dict[self._direction] + ' '
            except:
                raise ValueError('Constraint has no direction')
        if self._range == 0:
            s += ' {}'.format(- self._linCoef['CONST']['val'])
        else:
            s += ' [{}, {}]'.format(- self._linCoef['CONST']['val'],
                                    - self._linCoef['CONST']['val'] +
                                    self._range)
        return s

    def __repr__(self):
        """
        Returns a string representation of the object.
        """
        if self._name is not None:
            st = 'sasoptpy.Constraint({}, name=\'{}\')'
            s = st.format(str(self), self._name)
        else:
            st = 'sasoptpy.Constraint({}, name=None)'
            s = st.format(str(self))
        return s



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit test for solve statements.
"""

import os
import sys
import unittest
import warnings
import sasoptpy as so
from inspect import cleandoc

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '../..')))
from util import assert_equal_wo_temps

from sasoptpy.actions import solve, print_item

from tests.swat_config import create_cas_connection



class TestSolve(unittest.TestCase):
    """
    Unit tests for SOLVE statements
    """

    @classmethod
    def setUpClass(cls):
        so.reset()
        cls.conn = None
        from swat import CAS, SWATError
        try:
            cls.conn = create_cas_connection()
        except SWATError:
            warnings.warn('CAS connection is not available', RuntimeWarning)
        except TypeError:
            warnings.warn('CAS variables are not available', RuntimeWarning)

    @classmethod
    def tearDownClass(cls):
        if cls.conn is not None:
            cls.conn.close()

    def setUp(self):
        so.reset()

    def test_regular_solve(self):
        with so.Workspace('w') as w:
            x = so.Variable(name='x', lb=1, ub=10)
            o = so.Objective(2*x, sense=so.maximize, name='obj')
            s = solve()
            p = print_item(x)

        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc('''
            proc optmodel;
               var x >= 1 <= 10;
               max obj = 2 * x;
               solve;
               print x;
            quit;'''))

        if TestSolve.conn:
            w.set_session(TestSolve.conn)
            w.submit()
            self.assertEqual(str(p.get_response()), cleandoc('''
            x
            
                  x
            0  10.0'''))
            self.assertEqual(str(s.get_response()['Problem Summary']), cleandoc('''
                Problem Summary
                
                                                Value
                Label                                
                Objective Sense          Maximization
                Objective Function                obj
                Objective Type                 Linear
                                                     
                Number of Variables                 1
                Bounded Above                       0
                Bounded Below                       0
                Bounded Below and Above             1
                Free                                0
                Fixed                               0
                                                     
                Number of Constraints               0
                                                     
                Constraint Coefficients             0'''))

    def test_with_model(self):

        def produce_order_error():
            with so.Workspace('w') as w:
                m = so.Model(name='m')
                x = so.Variable(name='x')
                m.solve()
                m.include(x)
                m.solve()

        self.assertRaises(ReferenceError, produce_order_error)

        so.reset()
        with so.Workspace('w') as w:
            x = so.Variable(name='x')
            m = so.Model(name='m')
            m.include(x)
            m.solve()

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               var x;
               problem m include x;
               use problem m;
               solve;
            quit;'''))

    def test_with_options(self):
        with so.Workspace('w') as w:
            solve()
            solve(options={'with': 'milp'})
            solve(options={'with': 'milp'}, primalin=True)
            solve(options={'with': 'milp', 'presolver': None, 'feastol': 1e-6,
                           'logfreq': 2, 'maxsols': 3, 'scale': 'automatic',
                           'restarts': None, 'cutmir': 'aggressive'})

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               solve;
               solve with milp;
               solve with milp / primalin;
               solve with milp / presolver=None feastol=1e-06 logfreq=2 maxsols=3 scale=automatic restarts=None cutmir=aggressive;
            quit;'''))


    def test_warning_without_abstract(self):

        def solve_without_abstract():
            m = so.Model(name='m1')
            x = m.add_variable(name='x')
            m.set_objective(2*x, name='o')
            solve()

        self.assertWarns(UserWarning, solve_without_abstract)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from inspect import cleandoc
import unittest
import pandas as pd
import sasoptpy as so
from sasoptpy import to_expression as exp

import sys
import os
current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '..')))
from util import assert_equal_wo_temps


class TestUserUtils(unittest.TestCase):

    def setUp(self):
        so.reset()

    def test_concat(self):
        x = so.Variable(name='x')
        e1 = 'col'
        S = so.Set(name='S')
        e2 = so.SetIterator(S, name='i')
        r = so.util.concat(e1, e2)
        self.assertEqual(exp(r), '\'col\' || i')

    def test_exp_range(self):
        self.assertEqual(so.exp_range(1,5), range(1,5))
        t = so.exp_range(1, so.N)
        self.assertEqual(exp(t), '1.._N_')

    def test_get_value_table(self):

        self.assertEqual(so.get_value_table(), None)

        S = ['a', 'b', 'c']
        x = so.VariableGroup(S, name='x')
        x['a'].set_value(1)
        x['b'].set_value(3)
        x['c'].set_value(10)
        self.assertEqual(so.get_value_table(x).to_string(),
            cleandoc('''
                x
            a   1
            b   3
            c  10
            '''))

        c = so.ConstraintGroup((2 * x[i] >= 1 for i in S), name='c')
        self.assertEqual(so.get_value_table(c).to_string(), cleandoc(
            '''
                c
            a   2
            b   6
            c  20
            '''))

        df = pd.DataFrame([['a', 1, 2], ['b', 3, 4]],
                          columns=['var', 'lb', 'ub']).set_index(['var'])
        self.assertEqual(so.get_value_table(df).to_string(), cleandoc(
            '''
                lb ub
            var      
            a    1  2
            b    3  4
            '''))

        self.assertEqual(so.get_value_table(x, c).to_string(), cleandoc(
            '''
                x   c
            a   1   2
            b   3   6
            c  10  20
            '''
        ))

        self.assertEqual(so.get_value_table(df, df).to_string(), cleandoc(
            '''
                lb ub lb ub
            var            
            a    1  2  1  2
            b    3  4  3  4
            '''
        ))

        T = so.Set(name='T')
        y = so.VariableGroup(T, name='y')
        y[0].set_value(10)
        y[1].set_value(5)
        self.assertEqual(so.get_value_table(y).to_string(), cleandoc(
            '''
                y
            0  10
            1   5
            '''))

        z = so.ImplicitVar((2 * x[i] - 5 for i in S), name='z')
        self.assertEqual(so.get_value_table(z).to_string(), cleandoc(
            '''
                z
            a  -3
            b   1
            c  15
            '''
        ))

        e = so.Expression(x['a'] + 5 - y[0], name='e')
        self.assertEqual(so.get_value_table(e).to_string(), cleandoc(
            '''
               e
            - -4
            '''
        ))

    def test_submit(self):
        w = so.Workspace('w')
        def get_error_submit():
            w.submit()
        self.assertRaises(RuntimeError, get_error_submit)

        m = so.Model(name='m')
        def get_error_solve():
            m.solve()
        self.assertRaises(RuntimeError, get_error_solve)

    def test_expr_sum(self):
        S = ['a', 'b', 'c']
        x = so.VariableGroup(S, name='x')
        e = so.expr_sum(x[i] for i in S)

        T = so.Set(name='T')
        U = so.Set(name='U')
        y = so.VariableGroup(T, U, name='y')
        c = so.ConstraintGroup(
            (so.expr_sum(y[i, j] for i in T) <= 5 for j in U),
            name='c')
        assert_equal_wo_temps(
            self, so.to_definition(c),
            'con c {o10 in U} : sum {i in T} (y[i, o10]) <= 5;')

        e = so.expr_sum(y[i, 5] for i in T)
        assert_equal_wo_temps(self, exp(e), 'sum {i in T} (y[i, 5])')


    def test_reset_globals(self):
        def warn_reset_globals():
            so.reset_globals()
        self.assertWarns(DeprecationWarning, warn_reset_globals)

    def test_reset(self):
        x = so.Variable(name='x')
        self.assertGreater(so.itemid, 0)
        so.reset()
        self.assertEqual(so.itemid, 0)

    def test_dict_to_frame_and_flatten(self):
        d = {'coal': {'period1': 1, 'period2': 5, 'period3': 7},
           'steel': {'period1': 8, 'period2': 4, 'period3': 3},
           'copper': {'period1': 5, 'period2': 7, 'period3': 9}}
        df = so.dict_to_frame(d)
        self.assertEqual(df.to_string(), cleandoc(
            '''
                    period1  period2  period3
            coal          1        5        7
            steel         8        4        3
            copper        5        7        9
            '''))

        ff = so.flatten_frame(df)
        with pd.option_context('display.multi_sparse', False):
            self.assertEqual(ff.to_string(), cleandoc(
                '''
                coal    period1    1
                coal    period2    5
                coal    period3    7
                steel   period1    8
                steel   period2    4
                steel   period3    3
                copper  period1    5
                copper  period2    7
                copper  period3    9
                '''))
        ffs = so.flatten_frame(df, swap=True)
        self.assertEqual(ffs.to_string(), cleandoc(
            '''
            period1  coal      1
            period2  coal      5
            period3  coal      7
            period1  steel     8
            period2  steel     4
            period3  steel     3
            period1  copper    5
            period2  copper    7
            period3  copper    9
            '''))





#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from collections import OrderedDict
from math import copysign
import warnings

from sasoptpy.libs import np

import sasoptpy


class Expression:
    """
    Creates a mathematical expression to represent model components

    Parameters
    ----------
    exp : :class:`Expression`, optional
        An existing expression where arguments are being passed
    name : string, optional
        A local name for the expression

    Examples
    --------

    >>> x = so.Variable(name='x')
    >>> y = so.VariableGroup(3, name='y')
    >>> e = so.Expression(exp=x + 3 * y[0] - 5 * y[1], name='exp1')
    >>> print(e)
     -  5.0 * y[1]  +  3.0 * y[0]  +  x
    >>> print(repr(e))
    sasoptpy.Expression(exp =  -  5.0 * y[1]  +  3.0 * y[0]  +  x ,
                        name='exp1')

    >>> sales = so.Variable(name='sales')
    >>> material = so.Variable(name='material')
    >>> profit = 5 * sales - 3 * material
    >>> print(profit)
     5.0 * sales  -  3.0 * material
    >>> print(repr(profit))
    sasoptpy.Expression(exp =  5.0 * sales  -  3.0 * material , name=None)

    >>> import sasoptpy.abstract.math as sm
    >>> f = sm.sin(x) + sm.min(y[1],1) ** 2
    >>> print(type(f))
    <class 'sasoptpy.core.Expression'>
    >>> print(f)
    sin(x) + (min(y[1] , 1)) ** (2)

    Notes
    -----
    * Two other classes (:class:`Variable` and :class:`Constraint`) are
      subclasses of this class.
    * Expressions are created automatically after linear math operations
      with variables.
    * An expression object can be called when defining constraints and other
      expressions.
    """

    def __init__(self, exp=None, name=None):
        self._name = name
        if self._is_named_expression():
            self._objorder = sasoptpy.util.get_creation_id()
        self._value = 0
        self._dual = None
        self._linCoef = OrderedDict()
        if exp is None:
            self.set_member(key='CONST', ref=None, val=0)
        else:
            if isinstance(exp, Expression):
                self._copy_coef(exp)
            elif np.issubdtype(type(exp), np.number):
                self.set_member(key='CONST', ref=None, val=exp)
            else:
                raise TypeError('ERROR: Invalid type for expression: {}, {}'.format(exp, type(exp)))
        self._temp = False
        self._value = 0
        self._operator = None
        self._arguments = []
        self._iterkey = []
        self._abstract = False
        self.sym = sasoptpy.abstract.Conditional(self)
        self._keep = False

    @classmethod
    def to_expression(cls, obj):
        if sasoptpy.core.util.is_expression(obj):
            return obj
        else:
            if np.isinstance(type(obj), np.number):
                r = Expression(name=str(obj))
                r.add_to_member_value('CONST', obj)
                return r
            else:
                raise RuntimeError(
                    'Cannot convert type {} to Expression'.format(type(obj)))

    def copy(self, name=None):
        """
        Returns a copy of the :class:`Expression` object

        Parameters
        ----------
        name : string, optional
            Name for the copy

        Returns
        -------
        r : :class:`Expression`
            Copy of the object

        Examples
        --------

        >>> x = so.Variable(name='x')
        >>> y = so.VariableGroup(1, name='y')
        >>> e = so.Expression(7 * x - y[0], name='e')
        >>> print(repr(e))
        sasoptpy.Expression(exp =  -  y[0]  +  7.0 * x , name='e')
        >>> f = e.copy(name='f')
        >>> print(repr(f))
        sasoptpy.Expression(exp =  -  y[0]  +  7.0 * x , name='f')

        """
        r = Expression(name=name)
        for mylc in self._linCoef:
            r.copy_member(mylc, self)
        r._operator = self._operator
        r._iterkey = self._iterkey
        r._abstract = self._abstract
        r.sym.copy_conditions(self.sym)
        return r

    def _copy_coef(self, exp):
        for mylc in exp.get_member_dict():
            self.copy_member(mylc, exp)

    def _is_named_expression(self):
        return self._name is not None

    def get_member_dict(self):
        """
        Returns an ordered dictionary of elements
        """
        return self._linCoef

    def get_member(self, key):
        """
        Returns the requested member of the expression

        Parameters
        ----------
        key : string
            Identifier of the member, name for single objects

        Returns
        -------
        member : dict
            A dictionary of coefficient, operator, and reference of member
        """
        return self._linCoef.get(key, None)

    def set_member(self, key, ref, val, op=None):
        """
        Adds a new member or changes an existing member

        Parameters
        ----------
        key : string
            Identifier of the new or existing member
        ref : Object
            A reference to the new or existing member
        val : float
            Initial coefficient of the new or existing member
        op : string, optional
            Operator, if member has multiple children
        """
        self._linCoef[key] = {'ref': ref, 'val': val}
        if op:
            self._linCoef[key]['op'] = op

    def delete_member(self, key):
        """
        Deletes the requested member from the core dictionary
        """
        self._linCoef.pop(key, None)

    def copy_member(self, key, exp):
        """
        Copies the member of another expression

        Parameters
        ----------
        key : string
            Identifier of the member
        exp : :class:`Expression`
            Other expression to be copied from

        """
        self._linCoef[key] = dict(exp.get_member(key))

    def get_value(self):
        """
        Calculates and returns the value of the linear expression

        Returns
        -------
        v : float
            Value of the expression

        Examples
        --------

        >>> sales = so.Variable(name='sales', init=10)
        >>> material = so.Variable(name='material', init=3)
        >>> profit = so.Expression(5 * sales - 3 * material)
        >>> print(profit.get_value())
        41

        """
        v = 0
        for mylc in self.get_member_dict():
            if self.get_member(mylc)['ref'] is not None:
                if isinstance(mylc, tuple):
                    v += sasoptpy.core.util._evaluate(self.get_member(mylc))
                else:
                    m = self.get_member(mylc)
                    v += m['val'] * m['ref'].get_value()
            else:
                v += self.get_member(mylc)['val']
        if self._operator:
            try:
                import sasoptpy.abstract.math as sm
                if self._arguments:
                    vals = [i.get_value() if isinstance(i, Expression) else i for i in self._arguments]
                    v = sm.func_equivalent[self._operator](v, *vals)
                else:
                    v = sm.func_equivalent[self._operator](v)
            except KeyError:
                warnings.warn(
                    'Cannot evaluate operator: {} in {}'.format(
                        self._operator, self._expr()), RuntimeWarning)
        return v

    def get_member_value(self, key):
        """
        Returns coefficient of requested member

        Parameters
        ----------
        key : string
            Identifier of the member

        Returns
        -------
        value : float
            Coefficient value of the requested member
        """
        member = self.get_member(key)
        return member['val']

    def set_member_value(self, key, value):
        """
        Changes the coefficient of the requested member

        Parameters
        ----------
        key : string
            Identifier of the member
        value : float
            New coefficient value of the member
        """
        member = self.get_member(key)
        member['val'] = value

    def add_to_member_value(self, key, value):
        """
        Adds `value` to the coefficient of the requested member

        Parameters
        ----------
        key : string
            Identifier of the member
        value : float
            Value to be added
        """
        member = self.get_member(key)
        member['val'] += value

    def mult_member_value(self, key, value):
        """
        Multiplies the coefficient of the requested member by the specified
        `value`

        Parameters
        ----------
        key : string
            Identifier of the member
        value : float
            Value to be multiplied with
        """
        member = self.get_member(key)
        member['val'] *= value

    def get_dual(self):
        """
        Returns the dual value

        Returns
        -------
        dual : float
            Dual value of the object

        """
        return self._dual

    def set_dual(self, value):
        self._dual = value

    def set_name(self, name=None):
        """
        Specifies the name of the expression

        Parameters
        ----------
        name : string
            Name of the expression

        Returns
        -------
        name : string
            Name of the expression after resolving conflicts

        Examples
        --------

        >>> x = so.Variable(name='x')
        >>> e = x**2 + 2*x + 1
        >>> e.set_name('expansion')

        """
        if name:
            self._name = name
        else:
            self._name = 'o' + str(self._objorder)
        return self._name

    def get_name(self):
        """
        Returns the name of the object

        Returns
        -------
        name : string
            Name of the object

        Examples
        --------

        >>> m = so.Model()
        >>> var1 = m.add_variables(name='x')
        >>> print(var1.get_name())
        x
        """
        return self._name

    def get_name_with_keys(self, name=None):
        if name is None:
            name = self.get_name()

        if sasoptpy.core.util.is_key_empty(self._iterkey):
            return name
        else:
            return '{}{}'.format(
                name,
                sasoptpy.util.package_utils._to_optmodel_loop(self._iterkey))

    def set_temporary(self):
        """
        Converts expression into a temporary expression to enable in-place operations
        """
        self._temp = True

    def set_permanent(self):
        """
        Converts a temporary expression into a permanent one

        Parameters
        ----------
        name : string, optional
            Name of the expression

        Returns
        -------
        name : string
            Name of the expression in the namespace
        """
        self._temp = False
        if not sasoptpy.util.get_object_order(self):
            self._objorder = sasoptpy.util.get_creation_id()
        if self.get_name() is None:
            self.set_name()

    def get_constant(self):
        """
        Returns the constant term in the expression

        Examples
        --------

        >>> x = so.Variable(name='x')
        >>> e = 2 * x + 5
        >>> print(e.get_constant())
        5

        """
        return self.get_member('CONST')['val']

    def _expr_string(self, operator_type='sas'):
        # Add the operator first if exists
        s = ''
        if self._operator:
            s += self._operator
            if operator_type == 'sas':
                if self._iterkey:
                    if self._operator in sasoptpy.loop_operator_tags:
                        s += sasoptpy.util.package_utils._to_optmodel_loop(
                            self._iterkey, subindex=False) + ' '
            s += '('

        itemcnt = 0
        firstel = True

        # Add every element into string one by one
        for idx, el in self.get_member_dict().items():
            val = el['val']
            ref = el['ref']
            op = el.get('op')
            csign = copysign(1, val)

            # Skip elements with 0 coefficient or constant
            if val == 0 or idx == 'CONST':
                continue

            # Append sign of the value unless it is positive and first
            if firstel and csign == 1:
                pass
            elif csign == 1:
                s += '+ '
            elif csign == -1:
                s += '- '
            firstel = False

            # Add operand if exists, default is product
            if op:
                if operator_type == 'python':
                    optext = ' {} '.format(sasoptpy.util.get_python_symbol(op))
                else:
                    optext = ' {} '.format(op)
            else:
                optext = ' * '

            # For a list of expressions, a recursive function is called
            if isinstance(ref, list):
                if operator_type == 'python':
                    strlist = sasoptpy.util.package_utils._recursive_walk(
                        ref, func='__str__')
                else:
                    strlist = sasoptpy.util.package_utils._recursive_walk(
                        ref, func='_expr')
            elif sasoptpy.util.has_expr(ref):
                if operator_type == 'python':
                    strlist = [ref.__str__()]
                else:
                    strlist = [ref._expr()]
            else:
                strlist = [str(val)]

            if optext != ' * ' and optext != ' || ':
                strlist = ['({})'.format(stritem)
                           for stritem in strlist]

            # Merge all elements in strlist together
            refs = optext.join(strlist)
            if val == 1 or val == -1:
                if val == -1 and any(special_op in refs for special_op in
                                     sasoptpy.util.get_unsafe_operators()):
                    s += '({}) '.format(refs)
                else:
                    s += '{} '.format(refs)
            else:
                if op or any(special_op in refs for special_op in
                             sasoptpy.util.get_unsafe_operators()):
                    s += '{} * ({}) '.format(
                        sasoptpy.util.get_in_digit_format(abs(val)), refs)
                else:
                    s += '{} * {} '.format(
                        sasoptpy.util.get_in_digit_format(abs(val)), refs)

            itemcnt += 1

        # CONST is always at the end
        if itemcnt == 0 or (self.get_member_value('CONST') != 0 and
                            not sasoptpy.core.util.is_constraint(self)):
            val = self.get_member_value('CONST')
            csign = copysign(1, val)
            if csign < 0:
                s += '- '
            elif firstel:
                pass
            else:
                s += '+ '
            s += '{} '.format(sasoptpy.util.get_in_digit_format(abs(val)))

        # Close operator parentheses and add remaining elements
        if self._operator:
            if operator_type == 'python':
                if self._iterkey:
                    forlist = []
                    for i in self._iterkey:
                        key_expr = sasoptpy.core.util.get_key_for_expr(i)
                        if key_expr:
                            forlist.append(key_expr)
                    s += ' '.join(forlist)
            if self._arguments:
                if operator_type == 'sas':
                    s += ', ' + ', '.join(
                        i._expr() if hasattr(i, '_expr') else str(i) for i in
                        self._arguments)
                else:
                    s += ', ' + ', '.join(str(i) for i in self._arguments)
            s = s.rstrip()
            s += ')'
        s = s.rstrip()
        return s

    def _expr(self):
        """
        Generates the OPTMODEL-compatible string representation of the object

        Examples
        --------

        >>> x = so.Variable(name='x')
        >>> y = so.Variable(name='y')
        >>> f = x + y ** 2
        >>> print(f)
        x + (y) ** (2)
        >>> print(f._expr())
        x + (y) ^ (2)

        """
        return self._expr_string(operator_type='sas')


    def __repr__(self):
        """
        Returns a string representation of the object

        Examples
        --------

        >>> x = so.Variable(name='x')
        >>> y = so.Variable(name='y')
        >>> f = x + y ** 2
        >>> print(repr(f))
        sasoptpy.Expression(exp = x + (y) ** (2), name=None)

        """
        s = 'sasoptpy.Expression('
        if self._name is not None:
            s += 'exp = {}, name=\'{}\''.format(str(self), self._name)
        else:
            s += 'exp = {}, name=None'.format(str(self))
        s += ')'
        return s

    def __str__(self):
        """
        Generates a representation string that is Python-compatible

        Examples
        --------

        >>> f = x + y ** 2
        >>> print(str(f))
        x + (y) ** (2)

        """
        try:
            return self._expr_string(operator_type='python')
        except:
            return 'sasoptpy.Expression(id={})'.format(id(self))


    def _add_coef_value(self, var, key, value):
        """
        Changes value of a variable inside the :class:`Expression` object in
        place

        Parameters
        ----------
        var : :class:`Variable`
            Variable object whose value will be changed
        key : string
            Name of the variable object
        value : float
            New value or the addition to the existing value of the variable
        """
        if key in self.get_member_dict():
            self.set_member_value(key, value)
        else:
            self.set_member(key=key, ref=var, val=value)

    def add(self, other, sign=1):
        """
        Combines two expressions and produces a new one

        Parameters
        ----------
        other : float or :class:`Expression`
            Second expression or constant value to be added
        sign : int, optional
            Sign of the addition, 1 or -1

        Returns
        -------
        r : :class:`Expression`
            Reference to the outcome of the operation

        Notes
        -----
        * It is preferable to use regular Python operation, instead of calling
          this method:

          >>> e = x - y
          >>> f = 3 * x + 2 * y
          >>> g = e + f
          >>> print(g)
          4 * x + y

        """
        if self._temp and type(self) is Expression:
            r = self
        else:
            r = self.copy()

        if isinstance(other, Expression):
            r._abstract = self._abstract or other._abstract

            for v in other.get_member_dict():
                if r.get_member(v):
                    r.add_to_member_value(v, sign * other.get_member_value(v))
                else:
                    r.copy_member(v, other)
                    r.mult_member_value(v, sign)

            r.sym.copy_conditions(self.sym)
            r.sym.copy_conditions(other.sym)

        elif np.issubdtype(type(other), np.number):
            r.add_to_member_value('CONST', sign*other)
        else:
            raise TypeError(
                'Type {} for arithmetic operation is not valid.'.format(type(other)))

        return r

    def mult(self, other):
        """
        Multiplies the :class:`Expression` by a scalar value

        Parameters
        ----------
        other : :class:`Expression` or int
            Second expression to be multiplied

        Returns
        -------
        r : :class:`Expression`
            A new :class:`Expression` that represents the multiplication

        Notes
        -----
        * This method is mainly for internal use.
        * It is preferable to use regular Python operation, instead of calling
          this method:

          >>> e = 3 * (x-y)
          >>> f = 3
          >>> g = e*f
          >>> print(g)
          9 * x - 9 * y

        """
        if isinstance(other, Expression):
            r = Expression()
            r._abstract = self._abstract or other._abstract

            sasoptpy.core.util.multiply_coefficients(
                left=self.get_member_dict(),
                right=other.get_member_dict(),
                target=r.get_member_dict())

            r.sym.copy_conditions(self.sym)
            r.sym.copy_conditions(other.sym)

            return r
        elif np.issubdtype(type(other), np.number):
            if other == 0:
                r = Expression()
            else:
                r = self.copy()
                for e in r.get_member_dict():
                    r.mult_member_value(e, other)
            return r
        else:
            raise TypeError('Type for arithmetic operation is not valid.')

    def _relational(self, other, direction_):
        """
        Creates a logical relation between :class:`Expression` objects

        Parameters
        ----------
        other : :class:`Expression`
            Expression on the other side of the relation with respect to self
        direction_ : string
            Direction of the logical relation, either `E`, `L`, or `G`

        Returns
        -------
        generated_constraint : :class:`Constraint`
            Constraint generated as a result of linear relation

        """
        return sasoptpy.core.util.expression_to_constraint(left=self, relation=direction_, right=other)

    def _is_linear(self):
        """
        Checks whether the expression is composed of linear components

        Returns
        -------
        is_linear : boolean
            True if the expression is linear, False otherwise

        Examples
        --------

        >>> x = so.Variable()
        >>> e = x*x
        >>> print(e.is_linear())
        False

        >>> f = x*x + x*x - 2*x*x + 5
        >>> print(f.is_linear())
        True

        """

        # Loop over components
        for val in self.get_member_dict().values():
            if val.get('op', False):
                return False
            if type(val['ref']) is list:
                return False
            if val['ref'] and val['ref']._operator:
                return False
        return True

    def __hash__(self):
        return hash('{}{}'.format(self._name, id(self)))

    def __add__(self, other):
        return self.add(other)

    def __sub__(self, other):
        return self.add(other, -1)

    def __mul__(self, other):
        return self.mult(other)

    def __pow__(self, other):
        r = Expression()
        other = Expression.to_expression(other)
        self.set_permanent()
        other.set_permanent()
        key = (self.get_name(), other.get_name(), '^')
        ref = [self, other]
        val = 1
        op = '^'
        r.set_member(key=key, ref=ref, val=val, op=op)
        r._abstract = self._abstract or other._abstract
        return r

    def __rpow__(self, other):
        r = Expression()
        other = Expression.to_expression(other)
        self.set_permanent()
        other.set_permanent()
        key = (other.get_name(), self.get_name(), '^')
        ref = [other, self]
        val = 1
        op = '^'
        r.set_member(key=key, ref=ref, val=val, op=op)
        r._abstract = self._abstract or other._abstract
        return r

    def __radd__(self, other):
        return self.add(other)

    def __rsub__(self, other):
        tmp = self.add(other, -1)
        for v in tmp.get_member_dict():
            tmp.mult_member_value(v, -1)
        return tmp

    def __rmul__(self, other):
        return self.mult(other)

    def __truediv__(self, other):
        if np.issubdtype(type(other), np.number):
            try:
                return self.mult(1/other)
            except ZeroDivisionError:
                warnings.warn('Expression is divided by zero', RuntimeWarning)
                return None
        elif not isinstance(other, Expression):
            raise TypeError('Type for arithmetic operation is not valid.')

        r = Expression()
        other = Expression.to_expression(other)
        self.set_permanent()
        other.set_permanent()
        key = (self.get_name(), other.get_name(), '/')
        ref = [self, other]
        val = 1
        op = '/'
        r.set_member(key=key, ref=ref, val=val, op=op)
        r._abstract = self._abstract or other._abstract
        return r

    def __rtruediv__(self, other):
        r = Expression()
        other = Expression.to_expression(other)
        self.set_permanent()
        other.set_permanent()
        key = (other.get_name(), self.get_name(), '/')
        ref = [other, self]
        val = 1
        op = '/'
        r.set_member(key=key, ref=ref, val=val, op=op)
        r._abstract = self._abstract or other._abstract
        return r

    def concat(self, other):
        r = sasoptpy.Expression()
        key = (id(self), id(other), '||')
        ref = [self, other]
        val = 1
        op = '||'
        r.set_member(key=key, ref=ref, val=val, op=op)
        return r

    def __lt__(self, other):
        return self._relational(other, 'LT')

    def __gt__(self, other):
        return self._relational(other, 'GT')

    def __le__(self, other):
        return self._relational(other, 'L')

    def __ge__(self, other):
        return self._relational(other, 'G')

    def __eq__(self, other):
        return self._relational(other, 'E')

    def __ne__(self, other):
        return self._relational(other, 'NE')

    def __neg__(self):
        return self.mult(-1)

    def __iter__(self):
        return iter([self])


class Auxiliary(Expression):
    """
    Represents an auxiliary expression, often as a symbolic attribute

    Parameters
    ----------
    base : :class:`Expression`
        Original owner of the auxiliary value
    prefix : string, optional
        Prefix of the expression
    suffix : string, optional
        Suffix of the expression
    operator : string, optional
        Wrapping operator
    value : float, optional
        Initial value of the symbolic object

    Notes
    -----
    * Auxiliary objects are for internal use
    """

    def __init__(self, base, prefix=None, suffix=None, operator=None, value=None):
        super().__init__()
        self._base = base
        self._prefix = prefix
        self._suffix = suffix
        self._operator = operator
        self._value = value
        self.set_member(key=self.get_name(), ref=self, val=1)

    def get_prefix_str(self):
        if self._prefix is None:
            return ''
        return self._prefix

    def get_base_str(self):
        if self._base is None:
            return ''
        return sasoptpy.to_expression(self._base)

    def get_suffix_str(self):
        if self._suffix is None:
            return ''
        return self._suffix

    def wrap_operator_str(self, s):
        if self._operator is None:
            return s
        return '{}({})'.format(
            self._operator, s
        )

    def get_name(self):
        return self._expr()

    def _expr(self):
        prefix_str = self.get_prefix_str()
        base_str = self.get_base_str()
        suffix_str = self.get_suffix_str()
        s = prefix_str + base_str
        if suffix_str != '':
            s += '.' + suffix_str
        s = self.wrap_operator_str(s)
        return s


class Symbol(Expression):
    """
    Represents a symbolic string, to be evaluated on server-side

    Parameters
    ----------
    name : string
        String to be symbolized

    Notes
    -----
    * A Symbol object can be used for any values that does not translate to a value
      on client-side, but has meaning on execution. For example, `_N_` is a
      SAS symbol, which can be used in PROC OPTMODEL strings.
    """

    def __init__(self, name):
        super().__init__(name=name)
        self.set_member(key=self._name, ref=self, val=1)

    def _expr(self):
        return self._name

    def _defn(self):
        pass

    def __str__(self):
        return self._name



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from .statement_base import Statement
import sasoptpy


class Assignment(Statement):

    def __init__(self, identifier, expression, keyword=None):
        super().__init__()
        self.keyword = keyword
        self.identifier = identifier
        self.expression = expression

    def _defn(self):
        s = ''
        if self.keyword:
            s += self.keyword + ' '
        s += '{} = {};'.format(sasoptpy.to_expression(self.identifier),
                               sasoptpy.to_expression(self.expression))
        return s

    def append(self, *args, **kwargs):
        if kwargs.get('identifier'):
            self.identifier = kwargs.get('identifier')
        if kwargs.get('expression'):
            self.expression = kwargs.get('expression')
        if kwargs.get('keyword'):
            self.keyword = kwargs.get('keyword')

    @classmethod
    def set_bounds(cls, var, **kwargs):
        statements = []
        lb = kwargs.get('lb')
        ub = kwargs.get('ub')
        if lb and ub and lb is ub:
            from sasoptpy.abstract.statement import FixStatement
            statements.append(FixStatement((var, lb)))
        elif lb and ub:
            lb_identifier = sasoptpy.Symbol
            statements.append(Assignment(identifier=var.lb, expression=lb))
            statements.append(Assignment(identifier=var.ub, expression=ub))
        elif lb:
            statements.append(Assignment(identifier=var.lb, expression=lb))
        elif ub:
            statements.append(Assignment(identifier=var.ub, expression=ub))

        return statements

    @classmethod
    def set_value(cls, obj, value):
        st = Assignment(identifier=obj, expression=value)
        return st

    @classmethod
    def fix_value(cls, obj, value):
        st = Assignment(identifier=obj, expression=value, keyword='fix')
        return st




from math import inf
import warnings

import numpy as np
import pandas as pd

import sasoptpy

def to_mps(model, **kwargs):
    """
    Converts the Python model into a DataFrame object in MPS format

    Parameters
    ----------
    constant : boolean, optional
        Switching for using objConstant argument for solveMilp, solveLp. \
        Adds the constant as an auxiliary variable if value is True.

    Returns
    -------
    mpsdata : :class:`pandas.DataFrame`
        Problem representation in strict MPS format

    Examples
    --------

    >>> df = m.to_frame()
    >>> print(df)
         Field1 Field2  Field3 Field4 Field5 Field6 _id_
    0      NAME         model1      0             0    1
    1      ROWS                                        2
    2       MAX    obj                                 3
    3         L     c1                                 4
    4   COLUMNS                                        5
    5                x     obj      4                  6
    6                x      c1      3                  7
    7                y     obj     -5                  8
    8                y      c1      1                  9
    9       RHS                                       10
    10             RHS      c1      6                 11
    11   RANGES                                       12
    12   BOUNDS                                       13
    13   ENDATA                     0             0   14

    Notes
    -----
    * This method is called inside :meth:`Model.solve`.
    """

    constant = kwargs.get('constant', False)

    datarows = []

    def append_row(row):
        datarows.append(row + [len(datarows) + 1])
        return len(datarows)

    # Check if objective has a constant field
    if constant and model._objective._linCoef['CONST']['val'] != 0:
        obj_constant = model.add_variable('obj_constant')
        constant_value = model.get_objective().get_constant()
        obj_constant.set_bounds(lb=constant_value, ub=constant_value)
        obj_constant.set_value(constant_value)
        obj_name = model._objective.get_name() + '_constant'
        model.set_objective(model._objective - constant_value + obj_constant,
                           name=obj_name, sense=model._objective.get_sense())
        warnings.warn('WARNING: The objective function contains a'
                      ' constant term, an auxiliary variable is added.',
                      UserWarning)

    all_vars = model._get_all_variables()
    all_cons = model._get_all_constraints()

    # Create a dictionary of variables with constraint names
    var_con = {}
    for c in all_cons.values():
        for v in c._linCoef:
            var_con.setdefault(v, []).append(c.get_name())

    append_row(['NAME', '', model.get_name(), 0, '', 0])

    append_row(['ROWS', '', '', '', '', ''])
    if model._objective.get_name() is not None:
        append_row([model._objective._sense, model._objective.get_name(),
                          '', '', '', ''])
    for c in all_cons.values():
        append_row([c._direction, c.get_name(), '', '', '', ''])

    append_row(['COLUMNS', '', '', '', '', ''])
    curtype = sasoptpy.CONT
    for v in all_vars.values():
        f5 = 0
        if v._type is sasoptpy.INT and \
                curtype is sasoptpy.CONT:
            append_row(['', 'MARK0000', '\'MARKER\'', '',
                              '\'INTORG\'', ''])
            curtype = sasoptpy.INT
        if v._type is not sasoptpy.INT \
                and curtype is sasoptpy.INT:
            append_row(['', 'MARK0001', '\'MARKER\'', '',
                              '\'INTEND\'', ''])
            curtype = sasoptpy.CONT
        if v.get_name() in model._objective._linCoef:
            cv = model._objective._linCoef[v.get_name()]
            current_row = ['', v.get_name(), model._objective.get_name(), cv['val']]
            f5 = 1
        elif v.get_name() not in var_con:
            current_row = ['', v.get_name(), model._objective.get_name(), 0.0]
            f5 = 1
            var_con[v.get_name()] = []
        for cn in var_con.get(v.get_name(), []):
            if cn in all_cons:
                c = all_cons[cn]
                if v.get_name() in c._linCoef:
                    if f5 == 0:
                        current_row = ['', v.get_name(), c.get_name(),
                                       c._linCoef[v.get_name()]['val']]
                        f5 = 1
                    else:
                        current_row.append(c.get_name())
                        current_row.append(c._linCoef[v.get_name()]['val'])
                        ID = append_row(current_row)
                        f5 = 0
        if f5 == 1:
            current_row.append('')
            current_row.append('')
            ID = append_row(current_row)
    if curtype is sasoptpy.INT:
        append_row(['', 'MARK0001', '\'MARKER\'', '', '\'INTEND\'',
                          ''])

    append_row(['RHS', '', '', '', '', ''])
    f5 = 0
    for c in all_cons.values():
        if c._direction == 'L' and c._linCoef['CONST']['val'] == -inf:
            continue
        if c._direction == 'G' and c._linCoef['CONST']['val'] == 0:
            continue
        rhs = - c._linCoef['CONST']['val']
        if rhs != 0:
            if f5 == 0:
                current_row = ['', 'RHS', c.get_name(), rhs]
                f5 = 1
            else:
                current_row.append(c.get_name())
                current_row.append(rhs)
                f5 = 0
                append_row(current_row)
    if f5 == 1:
        current_row.append('')
        current_row.append('')
        append_row(current_row)

    append_row(['RANGES', '', '', '', '', ''])
    for c in all_cons.values():
        if c._range != 0:
            append_row(['', 'rng', c.get_name(), c._range, '', ''])

    append_row(['BOUNDS', '', '', '', '', ''])
    for v in all_vars.values():
        if v._lb == v._ub:
            append_row(['FX', 'BND', v.get_name(), v._ub, '', ''])
        if v._lb is not None and v._type is not sasoptpy.BIN:
            if v._ub == inf and v._lb == -inf:
                append_row(['FR', 'BND', v.get_name(), '', '', ''])
            elif not v._ub == v._lb:
                if v._type == sasoptpy.INT and \
                        v._lb == 0 and v._ub == inf:
                    append_row(['PL', 'BND', v.get_name(), '', '', ''])
                elif not (v._type == sasoptpy.CONT and v._lb == 0):
                    append_row(['LO', 'BND', v.get_name(), v._lb, '', ''])
        if v._ub != inf and v._ub is not None and not \
                (v._type is sasoptpy.BIN and v._ub == 1) and \
                v._lb != v._ub:
            append_row(['UP', 'BND', v.get_name(), v._ub, '', ''])
        if v._type is sasoptpy.BIN:
            append_row(['BV', 'BND', v.get_name(), '1.0', '', ''])

    append_row(['ENDATA', '', '', 0.0, '', 0.0])
    mpsdata = pd.DataFrame(data=datarows,
                           columns=['Field1', 'Field2', 'Field3', 'Field4',
                                    'Field5', 'Field6', '_id_'])
    datarows = []

    df = mpsdata
    for f in ['Field4', 'Field6']:
        df[f] = df[f].replace('', np.nan)
    df['_id_'] = df['_id_'].astype('int')
    df[['Field4', 'Field6']] = df[['Field4', 'Field6']].astype(float)

    return df



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit tests for math functions.
"""

import unittest
import warnings
from inspect import cleandoc

import sasoptpy as so
import sasoptpy.abstract.math as sm
import math


class TestAbstractMath(unittest.TestCase):
    """
    Unit tests for mathematical functions
    """

    def setUp(self):
        so.reset()

    def test_warn(self):
        def regular_import():
            import sasoptpy.math
        self.assertWarns(DeprecationWarning, regular_import)

    def test_functions(self):
        x = so.Variable(name='x')
        x.set_value(2)

        def check_value(expression, value, exp_str):
            self.assertEqual(expression.get_value(), value)
            self.assertEqual(so.to_expression(expression), exp_str)

        check_value(sm.abs(-x), 2, 'abs(- x)')
        check_value(sm.log(x/2), 0, 'log(0.5 * x)')
        check_value(sm.log2(2*x), 2, 'log2(2 * x)')
        check_value(sm.log10(5*x), 1, 'log10(5 * x)')
        check_value(sm.exp(x), math.exp(2), 'exp(x)')
        check_value(sm.sqrt(x**2), 2, 'sqrt((x) ^ (2))')
        check_value(sm.mod(5*x, 3), 1, 'mod(5 * x , 3)')
        check_value(sm.int(x+0.5), 2, 'int(x + 0.5)')
        check_value(sm.sign(-2 * x), -1, 'sign(- 2 * x)')
        check_value(sm.tan(x), math.tan(2), 'tan(x)')
        check_value(sm.sinh(x), math.sinh(2), 'sinh(x)')
        check_value(sm.cosh(x), math.cosh(2), 'cosh(x)')
        check_value(sm.tanh(x), math.tanh(2), 'tanh(x)')



import sasoptpy
from sasoptpy.core import Expression


class Objective(Expression):
    """
    Objective represents expressions with sense and used as target value in
    optimization

    Parameters
    ----------
    exp : :class:`Expression`
        Objective as an expression
    name : string
        Unique name of the expression
    sense : string, optional
        Direction of the objective, sasoptpy.MIN (default) or sasoptpy.MAX

    Examples
    --------

    >>> m = so.Model(name='test_objective')
    >>> x = m.add_variable(name='x')
    >>> obj = m.set_objective(2 * x - x ** 3, sense=so.MIN, name='new_obj')
    >>> str(m.get_objective())
    2 * x - (x) ** (3)
    >>> type(obj)
    sasoptpy.Objective

    """

    @sasoptpy.class_containable
    def __init__(self, exp, name, sense=None, **kwargs):
        super().__init__(exp=exp, name=name)
        if sense is None:
            sense = sasoptpy.MIN
        self._sense = sense
        self._default = kwargs.get('default', False)

    def _defn(self):
        return self._sense.lower() + ' ' + \
               self.get_name() + ' = ' + self._expr() + ';'

    def is_default(self):
        return self._default

    def get_sense(self):
        """
        Returns the objective sense (direction)
        """
        return self._sense

    def set_sense(self, sense):
        """
        Specifies the objective sense (direction)

        Parameters
        ----------
        sense : string
            sasoptpy.MIN or sasoptpy.MAX
        """
        self._sense = sense



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit tests for core classes.
"""

import unittest
import sasoptpy as so


class TestExpression(unittest.TestCase):
    """
    Unit tests for :class:`sasoptpy.Expression` objects
    """

    def setUp(self):
        pass

    def test_constructor(self):
        def unknown_expression_type():
            e = so.Expression(exp='abc')
        self.assertRaises(TypeError, unknown_expression_type)

    def test_get_value(self):
        import sasoptpy.abstract.math as sm
        x = so.Variable(name='x')
        x.set_value(7)
        e = sm.sin(2)**sm.sqrt(9) + sm.min(x, 5, 10)
        v = e.get_value()
        self.assertAlmostEqual(v, 5.751826, places=5)

    def test_dual(self):
        x = so.Variable(name='x')
        x._dual = 0
        self.assertEqual(x.get_dual(), 0)

    def test_rename(self):
        x = so.Variable(name='x')
        y = so.Variable(name='y')
        e = so.Expression(x + 2*x*y, name='nonlinear_exp')
        new_name = e.set_name('new_name')
        self.assertEqual(new_name, 'new_name')
        f = so.Expression(2*x + y)
        new_name = f.set_name('f')
        self.assertIsNotNone(new_name)
        new_name = e.set_name('e')
        self.assertEqual(new_name, 'e')
        get_name = e.get_name()
        self.assertEqual(get_name, 'e')

    def test_expr_string(self):
        import sasoptpy.abstract.math as sm
        x = so.Variable(name='x')
        y = so.Variable(name='y')
        e = sm.abs(x) + 2 *y + sm.min(x, y)
        e_exp = e._expr()
        self.assertEqual(e_exp, 'abs(x) + 2 * y + min(x , y)')

        setI = so.abstract.Set(name='I')
        z = so.VariableGroup(setI, name='z')
        g = so.expr_sum(z[i] for i in setI) + 5
        g_exp = g._expr()
        self.assertEqual(g_exp, 'sum {i in I} (z[i]) + 5')

    def test_repr(self):
        x = so.Variable(name='x')
        e = x ** 2 + 3 * x - 5
        exp_repr = repr(e)
        self.assertEqual(exp_repr, 'sasoptpy.Expression(exp = (x) ** (2) + 3 * x - 5, name=None)')
        e.set_permanent()
        e.set_name('e')
        exp_repr = repr(e)
        self.assertEqual(exp_repr, 'sasoptpy.Expression(exp = (x) ** (2) + 3 * x - 5, name=\'e\')')

    def test_string(self):
        x = so.Variable(name='x')
        e = 2 * x + 3 ** x + 10
        self.assertEqual(str(e), '2 * x + (3) ** (x) + 10')

        import sasoptpy.abstract.math as sm
        e = sm.abs(x) + sm.min(x, 2, sm.sqrt(x))
        self.assertEqual(str(e), 'abs(x) + min(x , 2, sqrt(x))')

        from sasoptpy.abstract import Set
        setI = Set(name='setI')
        y = so.VariableGroup(setI, name='y')
        e = - so.expr_sum(y[i] * i for i in setI)
        self.assertEqual(str(e), '- (sum(y[i] * i for i in setI))')

        e = 2 * x ** y[0]
        self.assertEqual(str(e), '2 * ((x) ** (y[0]))')

    def test_addition(self):
        x = so.Variable(name='x')
        y = so.Variable(name='y')

        exp1 = so.Expression(x+y, name='exp1')
        exp2 = 2 * x - y
        exp3 = exp1 + exp2
        self.assertEqual(str(exp3), '3 * x')

        import sasoptpy.abstract.math as sm
        exp4 = sm.min(x, y, sm.sqrt(x * y)) + 4
        self.assertEqual(str(exp4), 'min(x , y, sqrt(x * y)) + 4')

        def unknown_addition_type():
            exp5 = x + y + [3, 4]
        self.assertRaises(TypeError, unknown_addition_type)

    def test_multiplication(self):
        K = [1, 2, 'a']
        L = so.abstract.Set(name='L')

        x = so.Variable(name='x')
        y = so.Variable(name='y')
        z = so.VariableGroup(K, name='z')
        u = so.VariableGroup(L, name='u')

        exp1 = (x + y + 5) * (2 * x - 4 * y - 10)
        self.assertEqual(str(exp1), '2 * x * x - 2 * x * y - 4 * y * y - 30 * y - 50')

        exp2 = (x + y + 5) * 0 + x
        self.assertEqual(str(exp2), 'x')

        def unknown_multiplication_type():
            exp3 = x * ['a']
        self.assertRaises(TypeError, unknown_multiplication_type)

    def test_division(self):
        x = so.Variable(name='x')
        y = so.Variable(name='y')

        exp1 = x/2
        self.assertEqual(str(exp1), '0.5 * x')

        exp2 = x/y
        self.assertEqual(str(exp2), '(x) / (y)')

        import sasoptpy.abstract.math as sm
        exp3 = x**2 / (3*y) + 0 / sm.abs(x)
        self.assertEqual(str(exp3), '((x) ** (2)) / (3 * y) + (0) / (abs(x))')

        def division_by_zero():
            exp4 = x / 0
        self.assertWarns(RuntimeWarning, division_by_zero)

        def unknown_division_type():
           exp5 = x / 'a'
        self.assertRaises(TypeError, unknown_division_type)

    def test_power(self):
        x = so.Variable(name='x')
        e1 = x ** 2 + x ** 3
        self.assertEqual(so.to_expression(e1),
                         '(x) ^ (2) + (x) ^ (3)')

    def test_truediv(self):
        x = so.Variable(name='x')
        e1 = x / (2 + x) + x / (3 + x)
        self.assertEqual(so.to_expression(e1),
                         '(x) / (x + 2) + (x) / (x + 3)')

    def test_is_linear(self):
        x = so.Variable(name='x')
        y = so.Variable(name='y')

        exp1 = x
        self.assertTrue(exp1._is_linear())

        exp2 = 2 * x + 3 * y
        self.assertTrue(exp2._is_linear())

        exp3 = 5 * x * y
        self.assertFalse(exp3._is_linear())

        import sasoptpy.abstract.math as sm
        exp4 = sm.max(x, y) + 5
        self.assertFalse(exp4._is_linear())

        exp5 = x ** 2
        self.assertFalse(exp5._is_linear())

    def test_relational(self):
        x = so.Variable(name='x')
        e = 2 * x + 5

        f = e <= 2
        self.assertTrue(type(f) == so.Constraint)

        f = e >= 0
        self.assertTrue(type(f) == so.Constraint)

        f = e == 5
        self.assertTrue(type(f) == so.Constraint)

        f = 2 >= e
        self.assertTrue(type(f) == so.Constraint)

        f = 0 <= e
        self.assertTrue(type(f) == so.Constraint)

    def test_get_constant(self):
        x = so.Variable(name='x')

        e = 2 * x
        self.assertEqual(e.get_constant(), 0)

        e = 2 * x + 5
        self.assertEqual(e.get_constant(), 5)

    def tearDown(self):
        so.reset()




from collections import OrderedDict

from .parameter import Parameter

import sasoptpy

class ParameterGroup:
    """
    Represents a group of input parameters

    Parameters
    ----------
    index_key : iterable
        Index key of the group members
    name : string
        Name of the parameter group
    ptype : string, optional
        Type of the parameters. Possible values are `sasoptpy.STR` and
        `sasoptpy.NUM`
    value : float, optional
        Value of the parameter
    init : float, optional
        Initial value of the parameter

    Examples
    --------

    >>> from sasoptpy.actions import for_loop
    >>> with so.Workspace('w') as w:
    ...    p = so.ParameterGroup(so.exp_range(1, 6), name='p', init=3)
    ...    p[0].set_value(3)
    ...    S = so.Set(name='S', value=so.exp_range(1, 6))
    ...    for i in for_loop(S):
    ...        p[i].set_value(1)
    ...
    >>> print(so.to_optmodel(w))
    proc optmodel;
       num p {1..5} init 3;
       p[0] = 3;
       set S = 1..5;
       for {o13 in S} do;
          p[o13] = 1;
       end;
    quit;

    """

    @sasoptpy.class_containable
    def __init__(self, *index_key, name, init=None, value=None, ptype=None):
        self._key = list(index_key)
        self._name = name
        self._init = init
        self._value = value
        if ptype is None:
            ptype = sasoptpy.NUM
        self._ptype = ptype
        self._objorder = sasoptpy.util.get_creation_id()
        self._shadows = OrderedDict()

    def get_name(self):
        return self._name

    def get_element_name(self, key):
        keyname = sasoptpy.util.package_utils._to_sas_string(key)
        return '{}[{}]'.format(self.get_name(), keyname)

    def __getitem__(self, key):
        if key in self._shadows:
            return self._shadows[key]
        else:
            try:
                keys = ', '.join(i._expr()
                                 if hasattr(i, '_expr') else str(i) for i in
                                 key)
            except TypeError:
                keys = str(key)
            temp_name = '{}[{}]'.format(self.get_name(), keys)
            pv = Parameter(name=temp_name, internal=True)
            pv.set_parent(self, key=key)
            self._shadows[key] = pv
            return pv

    def __setitem__(self, key, value):
        k = self[key]
        k.set_value(value)
        sasoptpy.abstract.Assignment(self._shadows[key], value)

    def get_type_name_str(self):
        return '{} {}'.format(self._ptype, self.get_name())

    def get_iterator_str(self):
        key_defs = []
        for k in self._key:
            if sasoptpy.abstract.is_key_abstract(k):
                key_defs.append(k._defn())
            else:
                key_defs.append(sasoptpy.to_expression(k))
        return ' {{{}}}'.format(', '.join(key_defs))

    def get_value_init_str(self):
        if self._init is not None:
            return ' init {}'.format(self._init)
        elif self._value is not None:
            return ' = {}'.format(self._value)
        return ''

    def _defn(self):
        type_name_str = self.get_type_name_str()
        iterator_str = self.get_iterator_str()
        value_init_str = self.get_value_init_str()
        s = type_name_str + iterator_str + value_init_str
        s += ';'
        return s

    def _expr(self):
        return self.get_name()





import pandas as pd

Index = pd.Index
MultiIndex = pd.MultiIndex
Series = pd.Series
DataFrame = pd.DataFrame


def convert_dict_to_frame(dictobj, cols=None):
    df = pd.DataFrame.from_dict(dictobj, orient='index')
    if isinstance(cols, list):
        df.columns = cols
    if isinstance(df.index[0], tuple):
        df.index = pd.MultiIndex.from_tuples(df.index)
    return df


def display_dense():
    pd.set_option('display.multi_sparse', False)

def display_all():
    pd.set_option('display.max_columns', None)
    pd.set_option('display.max_rows', None)

concat = pd.concat





#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from .statement_base import Statement
import sasoptpy


class CreateDataStatement(Statement):

    def __init__(self, table, index, columns=None):
        super().__init__()
        self._table = table
        self._index = index
        if isinstance(index, list):
            self._index = {'key': index}
        self._columns = list()
        if columns:
            for col in columns:
                self.append(col)

    def append(self, element, **kwargs):
        if isinstance(element, dict):
            self._columns.append(element)
        else:
            self._columns.append({'expression': element})

    def get_table_expr(self):
        if hasattr(self._table, '_expr'):
            return f'({self._table._expr()})'
        elif hasattr(self._table, 'name'):
            return self._table.name
        else:
            return str(self._table)

    def get_table_name(self):
        return self._table

    def get_index_expr(self):
        s = ''
        key = self._index.get('key')
        if isinstance(key, str):
            key = [key]
        index = self._index.get('set')
        if key:
            joined_key = ' '.join([k.get_name() if hasattr(k, 'get_name') else str(k) for k in key])
            s += '[{}]'.format(joined_key)
        if key and index:
            s += ' = '
        if index:
            s += '{' + sasoptpy.util.package_utils._to_sas_string(index) + '}'
        return s

    def get_columns_expr(self):
        cols = self._columns
        columns_string = [CreateDataStatement.get_column_str(c) for c in cols]
        s = ' '.join(columns_string)
        return s

    @classmethod
    def get_column_str(cls, c):
        name = c.get('name')
        expr = c.get('expression', c.get('expr'))
        index = c.get('index')
        if not isinstance(index, list):
            index = [index]
        name_str = ''
        connect_str = ''
        expr_str = ''
        if name:
            if hasattr(name, '_expr'):
                name_str = 'col(' + name._expr() + ')'
            else:
                name_str = str(name)
        if name and expr:
            connect_str += '='
        if expr:
            if sasoptpy.util.has_expr(expr):
                expr_str += expr._expr()
            else:
                expr_str += '{}'.format(expr)

        if name_str == expr_str:
            connect_str = ''
            expr_str = ''

        if name_str != '':
            expr_str = '({})'.format(expr_str)

        s = name_str + connect_str + expr_str

        if index and any(sasoptpy.abstract.util.is_key_abstract(i) for i in index):
            index_str = ', '.join([sasoptpy.to_definition(i) for i in index])
            s = '{{{}}} < {} >'.format(index_str, s)

        return s

    def _defn(self):
        s = 'create data'
        table_str = self.get_table_expr()
        index_str = self.get_index_expr()
        column_str = self.get_columns_expr()

        s += ' ' + table_str + ' from '
        if index_str != '':
            s += index_str + ' '
        if column_str != '':
            s += column_str
        s += ';'
        return s

    @classmethod
    def create_data(cls, *args, **kwargs):
        c = CreateDataStatement(*args, **kwargs)
        return c




import sasoptpy.core.util
from sasoptpy.core.expression import Expression, Auxiliary, Symbol
from sasoptpy.core.objective import Objective
from sasoptpy.core.variable import Variable
from sasoptpy.core.group import Group
from sasoptpy.core.variable_group import VariableGroup
from sasoptpy.core.constraint import Constraint
from sasoptpy.core.constraint_group import ConstraintGroup
from sasoptpy.core.model import Model




import sasoptpy as so
import math


def test(cas_conn, num_guests=20, max_table_size=3, max_tables=None):

    m = so.Model("wedding", session=cas_conn)

    # Check max. tables
    if max_tables is None:
        max_tables = math.ceil(num_guests/max_table_size)

    # Sets
    guests = range(1, num_guests+1)
    tables = range(1, max_tables+1)
    guest_pairs = [[i, j] for i in guests for j in range(i+1, num_guests+1)]

    # Variables
    x = m.add_variables(guests, tables, vartype=so.BIN, name="x")
    unhappy = m.add_variables(tables, name="unhappy", lb=0)

    # Objective
    m.set_objective(unhappy.sum('*'), sense=so.MIN, name="obj")

    # Constraints
    m.add_constraints((x.sum(g, '*') == 1 for g in guests), name="assigncon")
    m.add_constraints((x.sum('*', t) <= max_table_size for t in tables),
                      name="tablesizecon")
    m.add_constraints((unhappy[t] >= abs(g-h)*(x[g, t] + x[h, t] - 1)
                       for t in tables for [g, h] in guest_pairs),
                      name="measurecon")

    # Solve
    res = m.solve(options={
        'with': 'milp', 'decomp': {'method': 'set'}, 'presolver': 'none'})

    if res is not None:

        print(so.get_solution_table(x))

        # Print assignments
        for t in tables:
            print('Table {} : [ '.format(t), end='')
            for g in guests:
                if x[g, t].get_value() == 1:
                    print('{} '.format(g), end='')
            print(']')

    return m.get_objective_value()



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Test for parameter class
"""

import os
import sys
import unittest
import warnings
import sasoptpy as so
from inspect import cleandoc

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '..')))
from util import assert_equal_wo_temps

class TestParameter(unittest.TestCase):

    def setUp(self):
        so.reset()

    def test_regular_parameter(self):

        with so.Workspace('w') as w:
            p = so.Parameter(name='p', init=3)
            p.set_value(5)

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               num p init 3;
               p = 5;
            quit;'''))

    def test_parameter_group(self):

        from sasoptpy.actions import for_loop, condition

        with so.Workspace('w') as w:
            p = so.ParameterGroup(so.exp_range(1, 6), name='p', init=3)
            p[0].set_value(3)
            S = so.Set(name='S', value=so.exp_range(1, 6))
            for i in for_loop(S):
                p[i].set_value(1)

        assert_equal_wo_temps(self, so.to_optmodel(w), cleandoc('''
            proc optmodel;
               num p {1..5} init 3;
               p[0] = 3;
               set S = 1..5;
               for {TEMP1 in S} do;
                  p[TEMP1] = 1;
               end;
            quit;'''))

    def test_value(self):

        p = so.Parameter(name='p', value=5)
        self.assertEqual(p.get_value(), 5)




from collections import OrderedDict
from types import GeneratorType

import sasoptpy
import sasoptpy.abstract
from .util import is_abstract
from sasoptpy.core.util import (is_expression)
from sasoptpy.util import to_expression, wrap_expression


class ImplicitVar:
    """
    Creates an implicit variable

    Parameters
    ----------
    argv : Generator, optional
        Generator object for the implicit variable
    name : string, optional
        Name of the implicit variable

    Notes
    -----

    - If the loop inside generator is over an abstract object, a definition
      for the object will be created inside :meth:`Model.to_optmodel` method.

    Examples
    --------

    Regular Implicit Variable

    >>> I = range(5)
    >>> x = so.Variable(name='x')
    >>> y = so.VariableGroup(I, name='y')
    >>> z = so.ImplicitVar((x + i * y[i] for i in I), name='z')
    >>> for i in z:
    >>>     print(i, z[i])
    (0,) x
    (1,) x + y[1]
    (2,) x + 2 * y[2]
    (3,) x + 3 * y[3]
    (4,) x + 4 * y[4]

    Abstract Implicit Variable

    >>> I = so.Set(name='I')
    >>> x = so.Variable(name='x')
    >>> y = so.VariableGroup(I, name='y')
    >>> z = so.ImplicitVar((x + i * y[i] for i in I), name='z')
    >>> print(z._defn())
    impvar z {i_1 in I} = x + i_1 * y[i_1];
    >>> for i in z:
    >>>     print(i, z[i])
    (sasoptpy.abstract.SetIterator(name=i_1, ...),) x + i_1 * y[i_1]

    """

    @sasoptpy.class_containable
    def __init__(self, argv, name=None):
        self._name = name
        self._objorder = sasoptpy.util.get_creation_id()
        self._dict = OrderedDict()
        self._conditions = []
        self._shadows = OrderedDict()
        if argv is not None:
            # Generator type - multi
            if type(argv) == GeneratorType:
                for arg in argv:
                    keynames = ()
                    keyrefs = ()
                    if argv.gi_code.co_nlocals == 1:
                        itlist = argv.gi_code.co_cellvars
                    else:
                        itlist = argv.gi_code.co_cellvars + argv.gi_code.co_varnames
                    localdict = argv.gi_frame.f_locals
                    for i in itlist:
                        if i != '.0':
                            keynames += (i,)
                    for i in keynames:
                        keyrefs += (localdict[i],)
                    self[keyrefs] = wrap_expression(arg)
            elif is_expression(argv):
                exp = wrap_expression(argv)
                exp.set_name(name)
                self[''] = exp
                self['']._objorder = self._objorder
            else:
                exp = sasoptpy.Expression.to_expression(argv)
                exp.set_name(name)
                self[''] = exp
                self['']._objorder = self._objorder

    def _expr(self):
        return self.get_name()

    def _defn(self):
        member_defn = []
        for i in self._dict.values():
            member_defn.append('impvar {} = {};'.format(
                i.get_name_with_keys(name=self.get_name()),
                to_expression(i)))
        s = '\n'.join(member_defn)
        return s

    def __setitem__(self, key, value):
        key = sasoptpy.util.pack_to_tuple(key)
        value.set_permanent()
        value._iterkey = key
        self._dict[key] = value

    def __getitem__(self, key):
        key = sasoptpy.util.pack_to_tuple(key)
        if key in self._dict:
            return self._dict[key]
        elif key in self._shadows:
            return self._shadows[key]
        else:
            tuple_key = sasoptpy.util.pack_to_tuple(key)
            pv = sasoptpy.abstract.ParameterValue(self, tuple_key)
            self._shadows[key] = pv
            return pv

    def get_name(self):
        return self._name

    def get_dict(self):
        return self._dict

    def get_keys(self):
        """
        Returns the dictionary keys

        Returns
        -------
        d : dict_keys
            Dictionary keys stored in the object
        """
        return self._dict.keys()

    def __iter__(self):
        return self._dict.__iter__()

    def __str__(self):
        return self._name








#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from inspect import cleandoc
import os
import unittest
import warnings

from swat import CAS, SWATError

import sasoptpy as so

from tests.swat_config import create_cas_connection


class Arbitrary:

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class MockCASResponse:

    def __init__(self, **kwargs):
        self.status = kwargs.get('status', 'OK')
        self.solutionStatus = kwargs.get('solutionStatus', 'OPTIMAL')

    def get_tables(self, *args):
        return [self.status]

    def keys(self):
        return list()


class MockCASServer:
    """
    Represents CAS Server for unit tests.
    """

    def __init__(self, **kwargs):
        so.util.get_session_type = self.get_session_type
        so.util.package_utils.get_session_type = self.get_session_type
        self.response = kwargs.get('response', MockCASResponse())
        self.optimization = Arbitrary()

    def get_session_type(self, *args, **kwargs):
        return 'CAS'

    def loadactionset(self, *args, **kwargs):
        pass

    def upload_frame(self, *args, **kwargs):
        return Arbitrary(name='name')

    def runOptmodel(self, *args, **kwargs):
        return self.response

    def solveLp(self, *args, **kwargs):
        return self.response

    def solveMilp(self, *args, **kwargs):
        return self.response


class TestCASInterface(unittest.TestCase):
    """
    Unit tests for the CAS interface
    """

    @classmethod
    def setUpClass(cls):
        so.reset()
        cls.conn = None
        try:
            cls.conn = create_cas_connection()
        except SWATError:
            warnings.warn('CAS connection is not available', RuntimeWarning)
        except TypeError:
            warnings.warn('CAS variables are not available', RuntimeWarning)

    @classmethod
    def tearDownClass(cls):
        if cls.conn is not None:
            cls.conn.close()

    def setUp(self):
        self.original_get_type = so.util.get_session_type
        self.original_package_get_type = so.util.package_utils.get_session_type

    def tearDown(self):
        so.util.get_session_type = self.original_get_type
        so.util.package_utils.get_session_type = self.original_package_get_type

    def test_unbounded(self):
        if not TestCASInterface.conn:
            self.skipTest('No session is available')

        m = so.Model(name='m', session=TestCASInterface.conn)
        x = m.add_variable(name='x', lb=0)
        m.set_objective(x, sense=so.MAX, name='obj')
        def warn_user():
            m.solve(frame=True, name='unbounded_test')
        self.assertWarns(UserWarning, warn_user)


    def test_forced_optmodel(self):
        if not TestCASInterface.conn:
            self.skipTest('No session is available')

        # Forced to OPTMODEL due abstract elements
        m = so.Model(name='test_forced_optmodel', session=TestCASInterface.conn)
        S = so.Set(name='S')
        m.include(S)
        x = m.add_variable(name='x', ub=1)
        m.set_objective(2*x, sense=so.MAX, name='obj')
        def warn_abstract():
            response = m.solve(mps=True, submit=False)
            self.assertIn('set S;', response)
        self.assertWarns(UserWarning, warn_abstract)

        # Uploaded MPS Table
        m.drop(S)
        response = m.solve(mps=True, submit=False)
        self.assertIn('CASTable', str(response))

        # Forced to OPTMODEL due nonlinearity
        m.set_objective(x**2, sense=so.MAX, name='obj2')
        def warn_nonlinear():
            response = m.solve(mps=True, submit=False)
            self.assertIn('var x', response)
        self.assertWarns(UserWarning, warn_nonlinear)

        # Forced to give error
        def force_decomp():
            m.solve(mps=True, options={'decomp': 'USER'})
        self.assertRaises(RuntimeError, force_decomp)

        # No objective
        m.solve(mps=True, verbose=True, name='no_obj')

    def test_regular_lp(self):
        if not TestCASInterface.conn:
            self.skipTest('No session is available')
        m = so.Model(name='test_regular_lp', session=TestCASInterface.conn)
        x = m.add_variable(name='x', lb=1)
        c = m.add_constraint(x <= 3, name='c')
        m.set_objective(x, sense=so.MIN, name='obj')
        m.solve(name='simple_lp')
        self.assertEqual(m.get_solution_summary().loc['Solver', 'Value'], 'LP')
        self.assertEqual(x.get_value(), 1)

    def test_primalin(self):
        if TestCASInterface.conn is None:
            self.skipTest('No session is available.')
        session = TestCASInterface.conn

        from sasoptpy import integer, minimize

        m = so.Model(name='test_primalin_mps', session=session)
        x = m.add_variable(name='x', vartype=integer, ub=9.5, lb=0)
        y = m.add_variable(name='y', vartype=integer, ub=5, lb=2)
        m.add_constraint(x + 2 * y >= 1, name='c1')
        m.add_constraint(x - 2 * y >= 2, name='c2')
        o = m.set_objective(x + y, sense=minimize, name='obj')
        m.solve(mps=True)

        self.assertEqual(so.get_value_table(x, y).to_string(), cleandoc('''
                 x    y
            -  6.0  2.0
            '''))

        m.set_objective(x + 1.1 * y, sense=minimize, name='obj2')
        m.solve(mps=True, primalin=True)
        self.assertEqual(
            session.CASTable('PRIMALINTABLE').to_frame().to_string(), cleandoc(
                '''
                Selected Rows from Table PRIMALINTABLE

                  _VAR_  _VALUE_
                0     x      6.0
                1     y      2.0
                '''
            ))

        # Drop option
        m.solve(mps=True, primalin=True, drop=True)
        def deleted_table():
            s = session.CASTable('PRIMALINTABLE').to_frame()
        self.assertRaises(SWATError, deleted_table)

    def test_response(self):
        if TestCASInterface.conn is None:
            self.skipTest('No session is available.')
        from sasoptpy import maximize, integer
        m = so.Model(name='test_response', session=TestCASInterface.conn)
        x = so.Variable(name='x', lb=1)
        o = so.Objective(x ** 2, sense=maximize, name='obj')
        m.include(x, o)
        m.solve(verbose=True)
        self.assertEqual(str(m.response.solutionStatus), 'UNBOUNDED')
        self.assertEqual(m._objval, None)

    def test_lp(self):
        if TestCASInterface.conn is None:
            self.skipTest('No session is available.')

        m = so.Model(name='model_lp', session=TestCASInterface.conn)
        x = m.add_variable(name='x', lb=1, ub=4)
        y = m.add_variable(name='y', lb=4, ub=10)
        m.add_constraint(x + y == 8, name='c1')
        m.set_objective(x-y, sense=so.MIN, name='obj')
        m.solve(mps=True, verbose=True)
        self.assertEqual(x.get_value(), 1)

    def test_workspace(self):
        if not TestCASInterface.conn:
            self.skipTest('No session is available')

        with so.Workspace('w', session=TestCASInterface.conn) as w:
            x = so.VariableGroup(3, name='x')
            c = so.ConstraintGroup((x[i] <= i for i in range(3)), name='c')
            o = so.Objective(x[1], name='obj', sense=so.MAX)
            s = so.actions.solve()

        self.assertEqual(so.to_optmodel(w), cleandoc('''
            proc optmodel;
               var x {{0,1,2}};
               con c_0 : x[0] <= 0;
               con c_1 : x[1] <= 1;
               con c_2 : x[2] <= 2;
               max obj = x[1];
               solve;
            quit;'''))
        r = w.submit()
        self.assertIsNotNone(s.get_solution_summary())
        self.assertIsNotNone(s.get_problem_summary())

        self.assertEqual(s.get_solution_summary().loc['Solver', 'Value'], 'LP')

        self.assertEqual(r.to_string(), cleandoc('''
        Selected Rows from Table SOLUTION
        
             i   var  value             lb             ub   rc
        0  1.0  x[0]    0.0 -1.797693e+308  1.797693e+308  0.0
        1  2.0  x[1]    1.0 -1.797693e+308  1.797693e+308  0.0
        2  3.0  x[2]    2.0 -1.797693e+308  1.797693e+308  0.0'''))

        self.assertEqual(x[1].get_value(), 1)

    def test_workspace_errors(self):

        mockResponse = MockCASResponse(status='Syntax Error')
        mockSession = MockCASServer(response=mockResponse)

        with so.Workspace('w', session=mockSession) as w:
            x = so.Variable(name='x')

        def produce_syntax_error():
            r = w.submit()
        self.assertRaises(SyntaxError, produce_syntax_error)

        mockResponse = MockCASResponse(status='Semantic Error')
        mockSession = MockCASServer(response=mockResponse)
        mockSession.optimization = Arbitrary(runoptmodel=True)
        m = so.Model(name='test_runtime', session=mockSession)
        m.add_variable(name='x')
        def produce_runtime_error():
            m.solve()
        self.assertRaises(RuntimeError, produce_runtime_error)

        with so.Workspace('w', session=mockSession) as w:
            x = so.Variable(name='x')
        def produce_runtime_error_for_ws():
            r = w.submit()
        self.assertRaises(RuntimeError, produce_runtime_error_for_ws)

    def test_errors_and_warnings(self):

        s = MockCASServer()
        m = so.Model(name='test_errors_for_cas', session=s)
        x = m.add_variable(name='x', lb=0)
        m.set_objective(x ** 2, name='obj', sense=so.MIN)

        def produce_runtime_error():
            m.solve(frame=False)
        self.assertRaises(RuntimeError, produce_runtime_error)

        m.set_objective(x, name='obj', sense=so.MIN)

        s = MockCASServer(response=MockCASResponse(status=['ERROR']))
        m.set_session(s)
        def produce_error_status():
            m.solve(frame=True)
        self.assertRaises(RuntimeError, produce_error_status)

    def test_tuner(self):

        import pandas as pd

        if TestCASInterface.conn is None:
            self.skipTest('No session is available')

        m = so.Model(name='knapsack_with_tuner', session=TestCASInterface.conn)
        data = [
            ['clock', 8, 4, 3],
            ['mug', 10, 6, 5],
            ['headphone', 15, 7, 2],
            ['book', 20, 12, 10],
            ['pen', 1, 1, 15]
        ]
        df = pd.DataFrame(data, columns=['item', 'value', 'weight',
                                         'limit']).set_index(['item'])
        ITEMS = df.index
        value = df['value']
        weight = df['weight']
        limit = df['limit']
        total_weight = 55
        # Variables
        get = m.add_variables(ITEMS, name='get', vartype=so.INT)
        # Constraints
        m.add_constraints((get[i] <= limit[i] for i in ITEMS), name='limit_con')
        m.add_constraint(
            so.expr_sum(weight[i] * get[i] for i in ITEMS) <= total_weight,
            name='weight_con')
        # Objective
        total_value = so.expr_sum(value[i] * get[i] for i in ITEMS)
        m.set_objective(total_value, name='total_value', sense=so.MAX)

        results = m.tune_parameters(tunerParameters={'maxConfigs': 10})

        self.assertEqual(len(results), 10)

        self.assertTrue('Configuration' in results.iloc[0].index)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from collections import OrderedDict
from math import inf

import sasoptpy
from sasoptpy.core import Expression


class Variable(Expression):
    """
    Creates an optimization variable to be used inside models

    Parameters
    ----------
    name : string
        Name of the variable
    vartype : string, optional
        Type of the variable
    lb : float, optional
        Lower bound of the variable
    ub : float, optional
        Upper bound of the variable
    init : float, optional
        Initial value of the variable
    abstract : boolean, optional
        When set to `True`, indicates that the variable is abstract

    Examples
    --------

    >>> x = so.Variable(name='x', lb=0, ub=20, vartype=so.CONT)
    >>> print(repr(x))
    sasoptpy.Variable(name='x', lb=0, ub=20, vartype='CONT')

    >>> y = so.Variable(name='y', init=1, vartype=so.INT)
    >>> print(repr(y))
    sasoptpy.Variable(name='y', lb=0, ub=inf, init=1, vartype='INT')

    See also
    --------
    :func:`sasoptpy.Model.add_variable`

    """

    @sasoptpy.class_containable
    def __init__(self, name, vartype=None, lb=None, ub=None,
                 init=None, abstract=False, key=None, **kwargs):
        super().__init__(name=name)
        if vartype is None:
            vartype = sasoptpy.CONT
        self._type = vartype

        self._lb, self._ub = sasoptpy.core.util.get_default_bounds_if_none(
            vartype, lb, ub)

        self._init = init
        if self._init is not None:
            self._value = self._init

        self._key = key
        self._parent = None
        self._temp = False
        self._abstract = abstract

        self._shadow = False
        self.sol = sasoptpy.core.Auxiliary(self, suffix='sol')
        self._initialize_self_coef()

    def _initialize_self_coef(self):
        self.set_member(key=self._name, ref=self, val=1)

    def _set_info(self, parent, key):
        self._parent = parent
        self._key = key

    def get_parent_reference(self):
        return self._parent, self._key

    @sasoptpy.containable
    def set_bounds(self, *, lb=None, ub=None):
        """
        Changes bounds on a variable

        Parameters
        ----------
        lb : float or :class:`Expression`
            Lower bound of the variable
        ub : float or :class:`Expression`
            Upper bound of the variable

        Examples
        --------

        >>> x = so.Variable(name='x', lb=0, ub=20)
        >>> print(repr(x))
        sasoptpy.Variable(name='x', lb=0, ub=20, vartype='CONT')
        >>> x.set_bounds(lb=5, ub=15)
        >>> print(repr(x))
        sasoptpy.Variable(name='x', lb=5, ub=15, vartype='CONT')

        """
        if lb is not None:
            self._lb = lb
        if ub is not None:
            self._ub = ub

    def set_init(self, init=None):
        """
        Changes initial value of a variable

        Parameters
        ----------
        init : float or None
            Initial value of the variable

        Examples
        --------

        >>> x = so.Variable(name='x')
        >>> x.set_init(5)

        >>> y = so.Variable(name='y', init=3)
        >>> y.set_init()

        """
        self._init = init

    def get_attributes(self):
        """
        Returns an ordered dictionary of main attributes

        Returns
        --------
        attributes : OrderedDict
            Dictionary consists of `init`, `lb`, and `ub` attributes

        """
        attributes = OrderedDict()
        attributes['init'] = self._init
        attributes['lb'] = self._lb
        attributes['ub'] = self._ub
        return attributes

    def get_name(self):
        if self._abstract or self._shadow:
            return str(self)
        else:
            return self._name

    @property
    def ub(self):
        """
        Upper bound of the variable
        """
        return sasoptpy.Auxiliary(self, suffix='ub', value=self._ub)

    @property
    def lb(self):
        """
        Lower bound of the variable
        """
        return sasoptpy.Auxiliary(self, suffix='lb', value=self._lb)

    def get_value(self):
        """
        Returns the value of the variable

        Returns
        -------
        value : float
            Value of the variable

        Examples
        --------

        >>> x.set_value(20)
        >>> x.get_value()
        20

        """
        if self._abstract:
            raise ValueError('Abstract variable cannot have a value.')
        return self._value

    @sasoptpy.containable
    def set_value(self, value):
        """
        Sets the value of a variable
        """
        self._value = value

    def set_parent(self, parent):
        self._parent = parent

    def get_type(self):
        """
        Returns the type of variable

        Valid values are:

        * `sasoptpy.CONT`
        * `sasoptpy.INT`
        * `sasoptpy.BIN`

        """
        return self._type

    def __repr__(self):
        """
        Returns a string representation of the object.
        """
        st = 'sasoptpy.Variable(name=\'{}\''.format(self.get_name())
        if self._lb != -inf:
            st += ', lb={}'.format(self._lb)
        if self._ub != inf:
            st += ', ub={}'.format(self._ub)
        if self._init is not None:
            st += ', init={}'.format(self._init)
        if self._abstract:
            st += ', abstract=True'
        st += ', vartype=\'{}\')'.format(self._type)
        return st

    def __str__(self):
        """
        Generates a representation string
        """
        if self._parent is not None and self._key is not None:
            keylist = [i._expr() if isinstance(i, Expression)
                       else str(i) for i in self._key]
            key = ', '.join(keylist)
            return ('{}[{}]'.format(self._parent._name, key))
        if self._shadow and self._iterkey:
            keylist = [i._expr() if isinstance(i, Expression)
                       else str(i) for i in self._iterkey]
            key = ', '.join(keylist)
            return('{}[{}]'.format(self._name, key))
        return '{}'.format(self._name)

    def _expr(self):
        if self._parent is not None and self._key is not None:
            keylist = sasoptpy.core.util._to_iterator_expression(self._key)
            key = ', '.join(keylist)
            return ('{}[{}]'.format(self._parent._name, key))
        if self._shadow and self._iterkey:
            keylist = sasoptpy.core.util.package_utils._to_iterator_expression(self._iterkey)
            key = ', '.join(keylist)
            return('{}[{}]'.format(self._name, key))
        return '{}'.format(self._name)

    def _defn(self):
        s = 'var {}'.format(self._name)
        BIN = sasoptpy.BIN
        CONT = sasoptpy.CONT
        if self._type != CONT:
            if self._type == BIN:
                s += ' binary'
            else:
                s += ' integer'
        if self._lb != -inf:
            if not (self._lb == 0 and self._type == BIN):
                s += ' >= {}'.format(self._lb)
        if self._ub != inf:
            if not (self._ub == 1 and self._type == BIN):
                s += ' <= {}'.format(self._ub)
        if self._init is not None:
            s += ' init {}'.format(self._init)
        s += ';'
        return(s)

    def _cond_expr(self):
        return self._expr() + '.sol'




from collections import OrderedDict
from types import GeneratorType

import sasoptpy
from sasoptpy.libs import pd
from sasoptpy.core import Group


class ConstraintGroup(Group):
    """
    Creates a group of :class:`Constraint` objects

    Parameters
    ----------
    argv : Generator-type object
        A Python generator that includes :class:`Expression` objects
    name : string, optional
        Name (prefix) of the constraints

    Examples
    --------

    >>> var_ind = ['a', 'b', 'c', 'd']
    >>> u = so.VariableGroup(var_ind, name='u')
    >>> t = so.Variable(name='t')
    >>> cg = so.ConstraintGroup((u[i] + 2 * t <= 5 for i in var_ind), name='cg')
    >>> print(cg)
    Constraint Group (cg) [
      [a:  2.0 * t  +  u['a']  <=  5]
      [b:  u['b']  +  2.0 * t  <=  5]
      [c:  2.0 * t  +  u['c']  <=  5]
      [d:  2.0 * t  +  u['d']  <=  5]
    ]

    >>> z = so.VariableGroup(2, ['a', 'b', 'c'], name='z', lb=0, ub=10)
    >>> cg2 = so.ConstraintGroup((2 * z[i, j] + 3 * z[i-1, j] >= 2 for i in
                                  [1] for j in ['a', 'b', 'c']), name='cg2')
    >>> print(cg2)
    Constraint Group (cg2) [
      [(1, 'a'):  3.0 * z[0, 'a']  +  2.0 * z[1, 'a']  >=  2]
      [(1, 'b'):  2.0 * z[1, 'b']  +  3.0 * z[0, 'b']  >=  2]
      [(1, 'c'):  2.0 * z[1, 'c']  +  3.0 * z[0, 'c']  >=  2]
    ]

    Notes
    -----
    Use :func:`sasoptpy.Model.add_constraints` when working with a single
    model.

    See also
    --------
    :func:`sasoptpy.Model.add_constraints`
    :func:`sasoptpy.Model.include`

    """

    @sasoptpy.class_containable
    def __init__(self, argv, name):
        super().__init__(name=name)
        self._condict = OrderedDict()
        self._shadows = OrderedDict()
        if type(argv) == list or type(argv) == GeneratorType:
            self._recursive_add_cons(argv, name=name, condict=self._condict)
            self.filter_unique_keys()
        elif argv is None:
            # Empty CG
            self._initialized = False
        else:
            raise(TypeError, "Invalid iterator type for constraint group")

        self._objorder = sasoptpy.util.get_creation_id()

    def get_name(self):
        """
        Returns the name of the constraint group

        Returns
        -------
        name : string
            Name of the constraint group

        Examples
        --------

        >>> m = so.Model(name='m')
        >>> x = m.add_variable(name='x')
        >>> indices = ['a', 'b', 'c']
        >>> y = m.add_variables(indices, name='y')
        >>> c1 = m.add_constraints((x + y[i] <= 4 for i in indices),
                                   name='con1')
        >>> print(c1.get_name())
        con1
        """
        return self._name

    def _get_name_list(self):
        def_names = []
        for i in self._condict.values():
            def_names.extend(i._get_name_list())
        return def_names

    def _recursive_add_cons(self, argv, name, condict, ckeys=()):
        conctr = 0

        for (idx, c) in enumerate(argv):

            if not sasoptpy.core.util.is_constraint(c):
                raise ValueError(
                    'Cannot create constraint from {}'.format(type(c)))

            if type(argv) == list:
                new_keys = ckeys + (idx,)
            elif type(argv) == GeneratorType:
                new_keys = sasoptpy.core.util.get_generator_names(argv)

            key_list = sasoptpy.core.util._to_safe_iterator_expression(new_keys)
            con_name = '{}[{}]'.format(name, ','.join(key_list))
            new_con = sasoptpy.Constraint(exp=c, name=con_name, crange=c._range)
            self[new_keys] = new_con
            conctr += 1

    def get_expressions(self, rhs=False):
        """
        Returns constraints as a list of expressions

        Parameters
        ----------
        rhs : boolean, optional
            When set to `True`, passes the the constant part (rhs) of the
            constraint

        Returns
        -------
        df : :class:`pandas.DataFrame`
            Returns a DataFrame that consists of constraints as expressions

        Examples
        --------

        >>> m = so.Model(name='m')
        >>> var_ind = ['a', 'b', 'c', 'd']
        >>> u = m.add_variables(var_ind, name='u')
        >>> t = m.add_variable(name='t')
        >>> cg = so.ConstraintGroup((u[i] + 2 * t <= 5 for i in var_ind),
                                    name='cg')
        >>> ce = cg.get_expressions()
        >>> print(ce)
                     cg
        a  u[a] + 2 * t
        b  u[b] + 2 * t
        c  u[c] + 2 * t
        d  u[d] + 2 * t
        >>> ce_rhs = cg.get_expressions(rhs=True)
        >>> print(ce_rhs)
                         cg
        a  u[a] + 2 * t - 5
        b  u[b] + 2 * t - 5
        c  u[c] + 2 * t - 5
        d  u[d] + 2 * t - 5

        """
        cd = OrderedDict()
        for i in self._condict:
            cd[i] = self._condict[i].copy()
            if rhs is False:
                cd[i]._linCoef['CONST']['val'] = 0
        cd_df = sasoptpy.util.dict_to_frame(cd, cols=[self._name])
        return cd_df

    def __getitem__(self, key):
        """
        Overloaded method to access individual constraints

        Parameters
        ----------
        key : string or int
            Key of the constraint

        Returns
        -------
        item : :class:`Constraint`
            Reference to the constraint
        """
        key = sasoptpy.util.pack_to_tuple(key)
        if key in self._condict:
            return self._condict[key]

        if key in self._shadows:
            return self._shadows[key]

        return self._get_shadow_if_abstract(key)

    def _get_shadow_if_abstract(self, keys):
        for i, k in enumerate(keys):
            group = self._groups[i]
            if any([sasoptpy.util.is_key_abstract(j) for j in group]):
                continue
            else:
                if k not in group:
                    return None
        # If it reached this part, it is a feasible request
        return self._create_shadow(keys)

    def __setitem__(self, key, value):
        tupled_key = sasoptpy.util.pack_to_tuple(key)
        self._condict[tupled_key] = value
        self._register_keys(tupled_key)
        value._set_info(parent=self, key=tupled_key)

    def _create_shadow(self, key):
        if len(self._condict) == 0:
            return None
        else:
            k = list(self._condict)[0]
            c = self._condict[k]
            cname = self.get_name()
            cname = cname.replace(' ', '')
            shadow = sasoptpy.Constraint(exp=c, direction=c._direction,
                                         name=cname, crange=c._range)
            self._shadows[key] = shadow
            shadow._set_info(parent=self, key=key, shadow=True)
            return shadow

    def __iter__(self):
        for i in self._condict.values():
            yield i
    #
    # def _set_con_info(self):
    #     for i in self._condict:
    #         self._condict[i]._set_info(parent=self, key=i)

    def get_members(self):
        """
        Returns a dictionary of members
        """
        return self._condict

    def get_member_by_name(self, name):
        keys = sasoptpy.abstract.util.get_key_from_name(name)
        return self[keys]

    def get_shadow_members(self):
        return self._shadows

    def get_all_keys(self):
        """
        Returns a list of all keys (indices) in the group
        """
        return list(self._condict.keys()) + list(self._shadows.keys())

    def _defn(self):
        from sasoptpy.util.package_utils import _to_optmodel_loop
        groups = []
        con_no = 0
        for key_ in self._condict:
            current_constraint = self._condict[key_]
            if sasoptpy.config['generic_naming']:
                con_name = self.get_name() + f'_{con_no}'
                con_no += 1
            else:
                con_name = self.get_name() + _to_optmodel_loop(key_, current_constraint)
            group_str = 'con {}'.format(con_name)
            group_str += ' : ' + self._condict[key_]._defn()
            group_str += ';'
            groups.append(group_str)
        return '\n'.join(groups)

    def _expr(self):
        return self.get_name()

    def __str__(self):
        """
        Generates a representation string
        """
        s = 'Constraint Group ({}) [\n'.format(self.get_name())
        for k in sorted(self._condict):
            v = self._condict[k]
            s += '  [{}: {}]\n'.format(sasoptpy.util.get_first_member(k), v)
        s += ']'
        return s

    def __repr__(self):
        """
        Returns a string representation of the object.
        """
        s = 'sasoptpy.ConstraintGroup(['
        s += ', '.join(str(self._condict[i]) for i in self._condict)
        s += '], '
        s += 'name=\'{}\')'.format(self.get_name())
        return s



# SAS MVA interface for sasoptpy

import sasoptpy
from sasoptpy.libs import np
from sasoptpy.interface import Mediator
from saspy import SASsession
from sasoptpy.interface.util import wrap_long_lines, replace_long_names

import warnings

class SASMediator(Mediator):
    """
    Handles the connection between sasoptpy and SAS instance

    Parameters
    ----------
    caller : :class:`sasoptpy.Model` or :class:`sasoptpy.Workspace`
        Model or workspace that mediator belongs to
    sas_session : :class:`saspy.SASsession`
        SAS session object

    Notes
    -----

    * SASMediator is used by :class:`sasoptpy.Model` and :class:`sasoptpy.Workspace` objects
      internally.

    """

    def __init__(self, caller, sas_session):

        self.caller = caller
        self.session = sas_session
        self.conversion = dict()

    def solve(self, **kwargs):
        """
        Solve action for :class:`Model` objects
        """
        mps_indicator = kwargs.get('mps', kwargs.get('frame', False))
        user_options = kwargs.get('options', dict())

        mps_indicator = self.is_mps_format_needed(mps_indicator, user_options)
        if mps_indicator:
            return self.solve_with_mps(**kwargs)
        else:
            return self.solve_with_optmodel(**kwargs)

    def submit(self, **kwargs):
        """
        Submit action for custom input and :class:`sasoptpy.Workspace` objects
        """
        return self.submit_optmodel_code(**kwargs)

    def is_mps_format_needed(self, mps_option, options):
        enforced = False
        session = self.session
        caller = self.caller

        model = caller

        if 'decomp' in options:
            mps_option = True
            enforced = True

        if mps_option:
            switch = False

            if model.get_sets() or model.get_parameters():
                warnings.warn(
                    'INFO: Model {} has abstract objects, '.format(
                        model.get_name()) + 'switching to OPTMODEL mode.',
                    UserWarning)
                switch = True
            elif not sasoptpy.is_linear(model):
                warnings.warn(
                    'INFO: Model {} include nonlinear components, '.format(
                        model.get_name()) + 'switching to OPTMODEL mode.',
                    UserWarning)
                switch = True

            if switch and mps_option and enforced:
                raise RuntimeError('Cannot run either in OPTMODEL or MPS mode.')
            elif switch:
                mps_option = False

        return mps_option

    def solve_with_mps(self, **kwargs):
        """
        Submits the problem in MPS (DataFrame) format, supported by old versions

        Parameters
        ----------
        kwargs : dict
            Keyword arguments for solver settings and options

        Returns
        -------
        primal_solution : :class:`pandas.DataFrame`
            Solution of the model or None

        """
        session = self.session
        model = self.caller

        verbose = kwargs.get('verbose', False)
        submit = kwargs.get('submit', True)
        #name = kwargs.get('name', None)
        name = sasoptpy.util.get_next_name()

        # Get the MPS data
        df = model.to_mps(constant=True)

        if verbose:
            print(df.to_string())

        if not submit:
            return df

        # Upload MPS table with new arguments
        try:
            session.df2sd(df=df, table=name, keep_outer_quotes=True)
        except TypeError:
            # If user is using an old version of saspy, apply the hack
            session.df2sd(df=df, table=name)
            session.submit("""
                        data {};
                            set {};
                            field3=tranwrd(field3, "'MARKER'", "MARKER");
                            field3=tranwrd(field3, "MARKER", "'MARKER'");
                            field5=tranwrd(field5, "'INTORG'", "INTORG");
                            field5=tranwrd(field5, "INTORG", "'INTORG'");
                            field5=tranwrd(field5, "'INTEND'", "INTEND");
                            field5=tranwrd(field5, "INTEND", "'INTEND'");
                        run;
                        """.format(name, name))

        # Find problem type and initial values
        ptype = 1  # LP
        for v in model.get_grouped_variables().values():
            if v._type != sasoptpy.CONT:
                ptype = 2
                break

        if ptype == 1:
            c = session.submit("""
                        ods output SolutionSummary=SOL_SUMMARY ProblemSummary=PROB_SUMMARY;
                        proc optlp data = {}
                           primalout  = solution
                           dualout    = dual;
                        run;
                        """.format(name))
        else:
            c = session.submit("""
                        ods output SolutionSummary=SOL_SUMMARY ProblemSummary=PROB_SUMMARY;
                        proc optmilp data = {}
                           primalout  = solution
                           dualout    = dual;
                        run;
                        """.format(name))

        logs = c['LOG']
        for line in logs.split('\n'):
            if not line[0:1].isdigit():
                print(line)

        return self.parse_sas_mps_solution()

    def solve_with_optmodel(self, **kwargs):
        """
        Submits the problem in OPTMODEL format

        Parameters
        ----------
        kwargs : dict
            Keyword arguments for solver settings and options

        Returns
        -------
        primal_solution : :class:`pandas.DataFrame`
            Solution of the model or None

        """

        model = self.caller
        session = self.session
        verbose = kwargs.get('verbose', False)
        submit = kwargs.get('submit', True)

        print('NOTE: Converting model {} to OPTMODEL.'.format(
            model.get_name()))
        options = kwargs.get('options', dict())
        primalin = kwargs.get('primalin', False)

        optmodel_string = model.to_optmodel(header=True, options=options,
                                           ods=False, primalin=primalin,
                                           parse=True)

        self.conversion = dict()

        # Check if any object has a long name
        limit_names = kwargs.get('limit_names', False)
        if limit_names:
            optmodel_string, conversion = replace_long_names(optmodel_string)
            self.conversion.update(conversion)

        wrap_lines = kwargs.get('wrap_lines', False)
        if wrap_lines:
            max_length = kwargs.get('max_line_length', 30000)
            optmodel_string = wrap_long_lines(optmodel_string, max_length)

        if verbose:
            print(optmodel_string)
        if not submit:
            return optmodel_string

        print('NOTE: Submitting OPTMODEL code to SAS instance.')

        optmodel_string = 'ods output SolutionSummary=SOL_SUMMARY ProblemSummary=PROB_SUMMARY;\n' + \
                          optmodel_string

        response = session.submit(optmodel_string)

        model.response = response

        # Print output
        for line in response['LOG'].split('\n'):
            first_word = line[0:1]
            if not first_word.isdigit():
                print(line)
            if 'WARNING 524' in line:
                raise RuntimeError(
                    r'Some object names are truncated, '
                    r'try submitting with limit_names=True parameter')
            elif 'The submitted line exceeds maximum line length' in line:
                raise RuntimeError(
                    r'Some lines exceed maximum line length, '
                    r'try submitting with wrap_lines=True parameter')

        # Parse solution
        return self.parse_sas_solution()

    def parse_sas_mps_solution(self):
        """
        Parses MPS solution after `solve` and returns solution
        """

        caller = self.caller
        session = self.session
        response = caller.response

        caller._problemSummary = self.parse_sas_table('PROB_SUMMARY')
        caller._solutionSummary = self.parse_sas_table('SOL_SUMMARY')

        solver = caller._solutionSummary.loc['Solver', 'Value']

        # Parse solution
        solution_df = session.sd2df('solution', libref='WORK')
        primalsoln = solution_df[['_VAR_', '_VALUE_', '_LBOUND_', '_UBOUND_']].copy()
        primalsoln.columns = ['var', 'value', 'lb', 'ub']
        if solver == 'LP':
            primalsoln['rc'] = solution_df['_R_COST_']
        caller._primalSolution = primalsoln

        dual_df = session.sd2df('dual', libref='WORK')
        dualsoln = dual_df[['_ROW_', '_ACTIVITY_']].copy()
        dualsoln.columns = ['con', 'value']
        if solver == 'LP':
            dualsoln['dual'] = dual_df['_VALUE_']
        caller._dualSolution = dualsoln

        caller._status = caller._solutionSummary.loc['Solution Status'].Value
        caller._soltime = float(
            caller._solutionSummary.loc['Solution Time'].Value)

        self.perform_postsolve_operations()
        return caller._primalSolution

    def parse_sas_solution(self):
        """
        Performs post-solve operations

        Returns
        -------
        solution : :class:`pandas.DataFrame`
         Solution of the problem
        """
        caller = self.caller
        session = self.session

        # Parse solution
        caller._primalSolution = session.sd2df('SOLUTION', libref='WORK')
        self.convert_to_original(caller._primalSolution)
        caller._dualSolution = session.sd2df('DUAL', libref='WORK')
        self.convert_to_original(caller._dualSolution)

        caller._problemSummary = self.parse_sas_table('PROB_SUMMARY')
        caller._solutionSummary = self.parse_sas_table('SOL_SUMMARY')

        caller._status = caller._solutionSummary.loc['Solution Status'].Value
        caller._soltime = float(caller._solutionSummary.loc['Solution Time'].Value)

        self.perform_postsolve_operations()
        return caller._primalSolution

    def parse_table(self, table):
        session = self.session
        return session.sd2df(table)

    def parse_sas_table(self, table_name):
        """
        Converts requested table name into :class:`pandas.DataFrame`
        """
        session = self.session

        parsed_df = session.sd2df(table_name)[['Label1', 'cValue1']]
        parsed_df.replace(np.nan, '', inplace=True)
        parsed_df.columns = ['Label', 'Value']
        parsed_df = parsed_df.set_index(['Label'])
        return parsed_df

    def convert_to_original(self, table):
        """
        Converts variable names to their original format if a placeholder gets
        used
        """
        if len(self.conversion) == 0:
            return

        name_from = []
        name_to = []
        for i in self.conversion:
            name_from.append(r'\b' + i + r'\b')
            name_to.append(self.conversion[i])
        table.replace(name_from, name_to, inplace=True, regex=True)



    def perform_postsolve_operations(self):
        """
        Performs post-solve operations for proper output display
        """
        caller = self.caller
        response = caller.response
        solution = caller._primalSolution
        dual = caller._dualSolution

        # Variable values
        solver = caller.get_solution_summary().loc['Solver', 'Value']
        for row in solution.itertuples():
            caller.set_variable_value(row.var, row.value)
            if solver == 'LP':
                caller.set_dual_value(row.var, row.rc)

        # Constraint values (dual) only for LP
        solver = caller.get_solution_summary().loc['Solver', 'Value']
        if solver == 'LP':
            for row in dual.itertuples():
                con = caller.get_constraint(row.con)
                if con is not None:
                    con.set_dual(row.dual)

        # Objective value
        if sasoptpy.core.util.is_model(caller):
            if 'Objective Value' in caller._solutionSummary.index:
                objval = caller._solutionSummary.loc['Objective Value'].Value
                objval = float(objval)
                caller.set_objective_value(objval)

        # Variable init values
        if sasoptpy.core.util.is_model(caller):
            for v in caller.loop_variables():
                v.set_init(v.get_value())

    def submit_optmodel_code(self, **kwargs):
        """
        Submits given :class:`sasoptpy.Workspace` object in OPTMODEL format

        Parameters
        ----------
        kwargs :
            Solver settings and options
        """

        caller = self.caller
        session = self.session
        optmodel_code = sasoptpy.util.to_optmodel(caller,
                                                  header=True,
                                                  parse=True,
                                                  ods=True)
        verbose = kwargs.get('verbose', None)

        # Check if any object has a long name
        limit_names = kwargs.get('limit_names', False)
        if limit_names:
            optmodel_code, conversion = replace_long_names(optmodel_code)
            self.conversion = conversion

        wrap_lines = kwargs.get('wrap_lines', False)
        if wrap_lines:
            max_length = kwargs.get('max_line_length', 30000)
            optmodel_code = wrap_long_lines(optmodel_code, max_length)

        if verbose:
            print(optmodel_code)

        response = session.submit(optmodel_code)

        caller.response = response

        # Print output
        for line in response['LOG'].split('\n'):
            first_word = line[0:1]
            if not first_word.isdigit():
                print(line)

        if session.SYSERR() != 0:
            raise RuntimeError('SAS submission failed with following error: {}'.
                               format(session.SYSERRORTEXT()))

        # caller.parse_solve_responses()
        # caller.parse_print_responses()

        # list of tables: session.list_tables('WORK')
        # soln: session.sd2df('SOLUTION', libref='WORK')
        # dual: session.sd2df('DUAL', libref='WORK')

        # Error msg: session.SYSERRORTEXT()
        # Error status: session.SYSERR()

        caller.parse_create_data_responses(self)
        return self.parse_sas_workspace_response()



    def parse_sas_workspace_response(self):
        """
        Parses results of workspace submission
        """
        caller = self.caller
        session = self.session
        response = caller.response

        solution = session.sd2df('SOLUTION', libref='WORK')
        self.convert_to_original(solution)
        dual_solution = session.sd2df('DUAL', libref='WORK')
        self.convert_to_original(dual_solution)

        caller._primalSolution = solution
        caller._dualSolution = dual_solution

        self.set_workspace_variable_values(solution)
        #self.set_constraint_values(dual_solution)

        # self.set_model_objective_value()
        # self.set_variable_init_values()

        return solution

    def set_workspace_variable_values(self, solution):
        """
        Performs post-solve assignment of :class:`sasoptpy.Workspace` variable values
        """
        caller = self.caller
        for row in solution.itertuples():
            caller.set_variable_value(row.var, row.value)



from .statement_base import Statement

import sasoptpy


class DropStatement(Statement):

    @sasoptpy.class_containable
    def __init__(self, *elements):
        super().__init__()
        for i in elements:
            self.elements.append(i)
        self.keyword = 'drop'

    def append(self, element):
        pass

    def _defn(self):
        s = f'{self.keyword} '
        cons = []
        for c in self.elements:
            cons.extend(c._get_name_list())
        s += ' '.join(cons) + ';'
        return s

    @classmethod
    def model_drop_constraint(cls, _, c):
        if sasoptpy.core.util.is_droppable(c):
            st = DropStatement(c)
            return st

    @classmethod
    def drop_constraint(cls, *constraints):
        if all([sasoptpy.core.util.is_droppable(c) for c in constraints]):
            st = DropStatement(*constraints)

class RestoreStatement(DropStatement):

    def __init__(self, *elements):
        super().__init__(*elements)
        self.keyword = 'restore'

    @classmethod
    def restore_constraint(cls, *constraints):
        if all([sasoptpy.core.util.is_droppable(c) for c in constraints]):
            st = RestoreStatement(*constraints)




from contextlib import contextmanager

from .statement_base import Statement

import sasoptpy


class ForLoopStatement(Statement):

    def __init__(self, *args):
        super().__init__()
        self.keyword = 'for'
        self._sets = list(args)

    @classmethod
    def for_loop(cls, *args):
        loop = ForLoopStatement(*args)
        return loop

    def __iter__(self):
        iterators = []
        for i in self._sets:
            j = sasoptpy.abstract.SetIterator(i)
            iterators.append(j)
        self.iterators = iterators
        self.original = sasoptpy.container

        sasoptpy.container = self
        if len(self.iterators) == 1:
            yield self.iterators[0]
        else:
            yield iter(self.iterators)
        sasoptpy.container = self.original

    def append(self, element):
        self.elements.append(element)

    def _defn(self):
        s = f'{self.keyword} '
        s += '{'
        loops = []
        for i, it in enumerate(self.iterators):
            loops.append('{} in {}'.format(
                sasoptpy.util.package_utils._to_sas_string(it),
                sasoptpy.util.package_utils._to_sas_string(self._sets[i])#._expr()
            ))
        s += ', '.join(loops)
        s += '} do;\n'

        eldefs = []
        for i in self.elements:
            eldef = sasoptpy.to_definition(i)
            if eldef is not None:
                eldefs.append(eldef)
        s += sasoptpy.util.addSpaces('\n'.join(eldefs), 3) + '\n'

        s += 'end;'
        return s



#!/bin/python3
# saspy configuration for local tests

import os

sshsas = {'saspath': os.environ.get('SASPATH'),
          'ssh': '/usr/bin/ssh',
          'host': os.environ.get('SASHOST'),
          'encoding': 'latin1',
          'options': ['-t', 'dev/mva-v940m8', '-box', 'laxno', '-nopp',
                      '-encoding', 'latin1'
                     ],
        #   'tunnel': 15005
          }
SAS_config_names = ['sshsas']



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

import unittest
import os
import sasoptpy
import sys
import time

current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '../../examples/client_side')))
sys.path.append(os.path.abspath(os.path.join(current_dir, '../../examples/server_side')))

from tests.swat_config import create_cas_connection


class NullWriter:

    def write(self, text): pass

    def flush(self): pass


class TestExamples(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.conn = None
        from swat import CAS, SWATError
        try:
            cls.conn = create_cas_connection()
            cls.defstdout = sys.stdout
        except SWATError:
            raise unittest.SkipTest('Cannot establish CAS connection. ' \
                  + 'Check your environment variables ' \
                  + '(CASHOST, CASPORT, AUTHINFO)')
        except TypeError:
            raise unittest.SkipTest('Environment variable may not be defined')

    @classmethod
    def tearDownClass(cls):
        if cls.conn is not None:
            cls.conn.close()

    def setUp(self):
        sasoptpy.config['max_digits'] = 12
        self.digits = 6

    def tearDown(self):
        sasoptpy.reset()

    def run_instance(self, test, **kwargs):
        #sys.stdout = NullWriter()
        t0 = time.time()
        val = test(TestExamples.conn, **kwargs)
        sys.stdout = TestExamples.defstdout
        if isinstance(val, tuple):
            print(test.__globals__['__file__'], val[0], time.time() - t0)
        else:
            print(test.__globals__['__file__'], val, time.time()-t0)
        return val

    def test_fm1(self):
        from food_manufacture_1 import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, 107842.5925925926, self.digits)

    def test_fm2(self):
        from food_manufacture_2 import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, 100278.7, 1)

    def test_fp1(self):
        from factory_planning_1 import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, 93715.17857142858, self.digits)

    def test_fp2(self):
        from factory_planning_2 import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, 108855.0, 1)

    def test_mp(self):
        from manpower_planning import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, 498677.2853185597, self.digits)

    def test_ro(self):
        from refinery_optimization import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, 211365.134768933, self.digits)

    def test_mo(self):
        from mining_optimization import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, 146.86, 2)

    def test_farmp(self):
        from farm_planning import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, 121719.17286133641, self.digits)

    def test_econ(self):
        from economic_planning import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, 2450.026622821294, self.digits)

    def test_decentral(self):
        from decentralization import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, 14.9, self.digits)

    def test_kidney_exchange(self):
        from sas_kidney_exchange import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, 17.11135898487, self.digits)

    def test_optimal_wedding(self):
        from sas_optimal_wedding import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, 13.0, self.digits)

    def test_curve_fitting(self):
        from curve_fitting import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, 1.475, self.digits)

    def test_nl1(self):
        from nonlinear_1 import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, 3.951157967716, 3)

    def test_nl2(self):
        from nonlinear_2 import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, -999.0, self.digits)

    def test_least_squares(self):
        from least_squares import test
        obj = self.run_instance(test)
        self.assertAlmostEqual(obj, 7.186296783293, self.digits)

    def test_efficiency_analysis(self):
        from efficiency_analysis import test
        obj, ed = self.run_instance(test, get_tables=True)
        en = ed.set_index('garage_name').loc['Woking', 'efficiency_number']
        self.assertAlmostEqual(en, 0.867320, self.digits)



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

from .statement_base import Statement
import sasoptpy


class ReadDataStatement(Statement):

    def __init__(self, table, index, columns=None):
        super().__init__()
        self._table = table
        self._index = index
        self._columns = list()
        if columns:
            for col in columns:
                self.append(col)

    def append(self, element, **kwargs):
        if isinstance(element, dict):
            self._columns.append(element)
        else:
            self._columns.append({'target': element})

    def get_table_expr(self):
        if hasattr(self._table, 'name'):
            return self._table.name
        else:
            return str(self._table)

    def get_index_expr(self):
        index = self._index
        target = index.get('target')
        key = index.get('key')

        s = ''
        if target:
            s += '{}'.format(index['target'])
        if target and key:
            s += '='
        if key:
            s += '[{}]'.format(
                ReadDataStatement.flatten_column(key))
        return s

    def get_columns_expr(self):
        cols = self._columns
        columns_string = [ReadDataStatement.get_column_str(c) for c in cols]
        s = ' '.join(columns_string)
        return s

    @classmethod
    def flatten_column(cls, col):
        if isinstance(col, list) or isinstance(col, sasoptpy.SetIteratorGroup):
            ind = [ReadDataStatement.flatten_column(c) for c in col]
            return ' '.join(ind)
        elif isinstance(col, str):
            return col
        else:
            return str(col)

    @classmethod
    def get_column_str(cls, c):
        target = c.get('target')
        column = c.get('column')
        index = c.get('index')
        target_str = ''
        connect_str = ''
        column_str = ''
        if target:
            target_str += '{}'.format(ReadDataStatement.get_target_expr(target))
        if target and column:
            connect_str += '='
        if column:
            if sasoptpy.util.has_expr(column):
                column_str += 'col(' + column._expr() + ')'
            else:
                column_str += '{}'.format(column)

        # If both equal, no need for second part
        if target_str == column_str:
            connect_str = ''
            column_str = ''

        s = target_str + connect_str + column_str

        if index and sasoptpy.abstract.util.is_key_abstract(index):
            s = '{{{}}} < {} >'.format(sasoptpy.to_definition(index), s)

        return s

    @classmethod
    def get_target_expr(cls, target):
        return sasoptpy.to_expression(target)

    def _defn(self):
        s = 'read data '
        table_str = self.get_table_expr()
        index_str = self.get_index_expr()
        col_str = self.get_columns_expr()

        s += table_str + ' into'
        if index_str != '':
            s += ' ' + index_str
        if self._columns is not None:
            s += ' ' + col_str
        s += ';'
        return s

    @classmethod
    def read_data(cls, *args, **kwargs):
        r = ReadDataStatement(*args, **kwargs)
        return r



#!/usr/bin/env python
# encoding: utf-8
#
# Copyright SAS Institute
#
#  Licensed under the Apache License, Version 2.0 (the License);
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#

"""
Unit tests for core classes.
"""

import unittest
import sasoptpy as so
import pandas as pd

import os
import sys
current_dir = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(current_dir, '..')))
from util import assert_equal_wo_temps


class TestVariableGroup(unittest.TestCase):
    """
    Unit tests for :class:`sasoptpy.VariableGroup` objects
    """

    def setUp(self):
        pass

    def test_constructor(self):
        idx = [1, 2, 3]
        v = so.VariableGroup(idx, name='vg_1')
        self.assertEqual(repr(v),
                         "sasoptpy.VariableGroup([1, 2, 3], name='vg_1')")

        u = so.VariableGroup(5, name='u')
        self.assertEqual(repr(u),
                         "sasoptpy.VariableGroup([0, 1, 2, 3, 4], name='u')")

        # w = so.VariableGroup([('a', 1), ('a', 2), ('b', 1), ('b', 2)], name='w')
        # self.assertEqual(repr(w),
        #                  "sasoptpy.VariableGroup(['a', 'b'], [1, 2], name='w')")

        def pass_error():
            z = so.VariableGroup(name='z')
        self.assertRaises(ValueError, pass_error)

    def test_get_name(self):
        # Test regular name
        u = so.VariableGroup(4, name='myvargroup')
        self.assertEqual(u.get_name(), 'myvargroup')

        # Test default name
        v = so.VariableGroup(5, name=None)
        self.assertEqual(v.get_name(), v._name)

    def test_abstract(self):
        u = so.VariableGroup(4, name='u')
        self.assertFalse(u._abstract)
        u.set_abstract(True)
        self.assertTrue(u._abstract)

        I = so.Set(name='I')
        v = so.VariableGroup(I, name='v')
        self.assertTrue(v._abstract)
        v.set_abstract(False)
        self.assertFalse(v._abstract)

    def test_get_item(self):
        I = so.Set(name='I')
        v = so.VariableGroup(I, name='v')

        for i in I:
            individual_variable = v[i]
            self.assertEqual(str(individual_variable), 'v[{}]'.format(i))
            same_variable = v[i]
            self.assertEqual(str(same_variable), 'v[{}]'.format(i))

        idx1 = [1, 2, 3]
        idx2 = ['a', 'b', 'c']
        x = so.VariableGroup(idx1, idx2, name='x')
        self.assertEqual(len(x[1,]), 3)
        self.assertEqual(type(x[1]), list)

        for e, i in enumerate(x['*', 'a']):
            self.assertEqual(str(i), 'x[{}, a]'.format(e+1))

        def empty_list():
            a = x['*', 'd']
        self.assertWarns(RuntimeWarning, empty_list)

    def test_defn(self):
        idx1 = [1, 2, 3]
        idx2 = ['a', 'b', 'c']
        x = so.VariableGroup(idx1, idx2, name='x')
        self.assertEqual(x._defn(), "var x {{1,2,3}, {'a','b','c'}};")

        from sasoptpy.abstract import Set, Parameter
        idx3 = Set(name='I')
        y = so.VariableGroup(idx3, name='y')
        self.assertEqual(y._defn(), "var y {{I}};")

        z = so.VariableGroup(4, vartype=so.INT, name='z', lb=1)
        self.assertEqual(z._defn(), "var z {{0,1,2,3}} integer >= 1;")
        z.set_bounds(ub=10)
        self.assertEqual(z._defn(), "var z {{0,1,2,3}} integer >= 1 <= 10;")
        z.set_init(5)
        self.assertEqual(z._defn(), "var z {{0,1,2,3}} integer >= 1 <= 10 init 5;")

        u = so.VariableGroup(5, vartype=so.BIN, name='u')
        self.assertEqual(u._defn(), "var u {{0,1,2,3,4}} binary;")

        t = so.VariableGroup(idx3, name='t', init=10)
        self.assertEqual(t._defn(), "var t {{I}} init 10;")

        #         w = so.VariableGroup(2, name='w', lb=0, ub=10)
        #         w[0].set_bounds(lb=2, ub=8)
        #         w[1].set_bounds(lb=1, ub=7)
        #         expected_def = """var w {{0,1}} >= 0 <= 10;
        # w[0].lb = 2;
        # w[0].ub = 8;
        # w[1].lb = 1;
        # w[1].ub = 7;"""
        #         self.assertEqual(w._defn(), expected_def)
        #         w[0].set_init(4)
        #         expected_def = """var w {{0,1}} >= 0 <= 10;
        # w[0].lb = 2;
        # w[0].ub = 8;
        # w[0] = 4;
        # w[1].lb = 1;
        # w[1].ub = 7;"""
        #         self.assertEqual(w._defn(), expected_def)

    def test_add_member(self):
        from sasoptpy.abstract import Set
        I = Set(name='I')
        x = so.VariableGroup(I, name='x')
        x.add_member(key=0, init=5)
        self.assertEqual(str(x[0]), "x[0]")

        y = so.Variable(name='y')
        x.include_member('y', var=y)
        self.assertEqual(id(y), id(x['y']))

    def test_sum(self):
        I = so.Set(name='I')
        x = so.VariableGroup(I, name='x')
        e = x.sum('*')
        key = e._iterkey[0]
        self.assertEqual(e._expr(), "sum {{{k} in I}} (x[{k}])".format(k=key))

        z = so.VariableGroup(3, 4, name='z')
        e = z.sum(0, (1, 2))
        self.assertEqual(e._expr(), "z[0, 1] + z[0, 2]")

        idx1 = ['a', 'b', 'c']
        y = so.VariableGroup(I, idx1, name='y')
        e = y.sum('*', 'a')
        key = e._iterkey[0]
        self.assertEqual(e._expr(), "sum {{{k} in I}} (y[{k}, 'a'])".format(
            k=key
        ))

        e = y.sum('*', ('a', 'b'))
        key = e._iterkey[0]
        self.assertEqual(
            e._expr(),
            "sum {{{k} in I}} (y[{k}, 'a'] + y[{k}, 'b'])".format(k=key))

        e = y.sum(1, '*')
        self.assertEqual(e._expr(), "y[1, 'a'] + y[1, 'b'] + y[1, 'c']")

        w = so.VariableGroup([1, 2, 3], ['a', 'b'], name='w')
        e = w.sum([1, 3], '*')
        self.assertEqual(e._expr(), "w[1, 'a'] + w[1, 'b'] + w[3, 'a'] + w[3, 'b']")

    def test_mult(self):
        x = so.VariableGroup(3, name='x')
        m1 = [2, 3, 4]
        e = x.mult(m1)
        self.assertEqual(e._expr(), "2 * x[0] + 3 * x[1] + 4 * x[2]")

        m2 = pd.Series(m1)
        e = x.mult(m2)
        self.assertEqual(e._expr(), "2 * x[0] + 3 * x[1] + 4 * x[2]")

        data = [['a', 1, 3, 7], ['b', 4, 5, 6]]
        df = pd.DataFrame(
            data, columns=['idx', 'low', 'med', 'high']).set_index(['idx'])
        y = so.VariableGroup(df.index, df.columns, name='y')
        e = y.mult(df)
        self.assertEqual(
            e._expr(),
            "y['a', 'low'] + 3 * y['a', 'med'] + 7 * y['a', 'high'] + 4 * y['b', 'low'] + 5 * y['b', 'med'] + 6 * y['b', 'high']"
        )

        idx = ['a', 'b', 'c']
        z = so.VariableGroup(idx, name='z')
        data_dict = {'a': 1, 'b': 2, 'c': -4}
        e = z.mult(data_dict)
        self.assertEqual(e._expr(), "z['a'] + 2 * z['b'] - 4 * z['c']")

        idx = [1, 2]
        w = so.VariableGroup(idx, idx, name='w')
        data_dict = {(1, 1): 3, (1,2): 4, (2,1): 7, (2,2): 2}
        e = w.mult(data_dict)
        self.assertEqual(
            e._expr(), "3 * w[1, 1] + 4 * w[1, 2] + 7 * w[2, 1] + 2 * w[2, 2]")

        def unknown_key():
            data_dict = {(1, 1): 3, (3, 3): 9}
            e = w.mult(data_dict)
            e._expr()
        self.assertRaises(KeyError, unknown_key)

    def test_init(self):
        x = so.VariableGroup(3, name='x', init=4)
        self.assertTrue(all(i._init == 4 for i in x))
        x.set_init(5)
        self.assertTrue(all(i._init == 5 for i in x))

        from sasoptpy.abstract import Set
        I = Set(name='I')
        y = so.VariableGroup(I, name='y')
        y[0].set_init(5)
        self.assertEqual(y[0]._init, 5)
        y.set_init(4)
        self.assertEqual(y[0]._init, 4)

    def test_str(self):
        x = so.VariableGroup(2, name='x')
        self.assertEqual(
            str(x), "Variable Group (x) [\n  [0: x[0]]\n  [1: x[1]]\n]"
        )

    def test_set_bounds(self):
        x = so.VariableGroup(2, name='x')
        x.set_bounds(ub=2)
        self.assertEqual(x._ub, 2)
        self.assertEqual(so.to_definition(x), "var x {{0,1}} <= 2;")
        x.set_bounds(lb=1)
        self.assertEqual(x._lb, 1)
        self.assertEqual(so.to_definition(x), "var x {{0,1}} >= 1 <= 2;")
        x[0].set_bounds(lb=5)
        self.assertEqual(so.to_definition(x), "var x {{0,1}} >= 1 <= 2;")
        self.assertEqual(x._member_defn(), "x[0].lb = 5;")
        x[1].set_bounds(ub=10)
        self.assertEqual(x._member_defn(), "x[0].lb = 5;\nx[1].ub = 10;")

    def test_shadow(self):
        I = so.Set(name='I')
        x = so.VariableGroup(I, name='x')
        j = so.SetIterator(I, name='j')
        self.assertEqual(so.to_expression(x[j]), 'x[j]')
        y = so.VariableGroup(I, I, name='y')

        from sasoptpy.util import iterate

        with iterate([I, I], name='i') as i:
            z = y[i]
            self.assertEqual(so.to_expression(z), 'y[i]')

        J = so.Set(name='J', settype=[so.string, so.string])
        w = so.VariableGroup(J, name='w')
        for j in J:
            assert_equal_wo_temps(self, so.to_expression(w[j]), 'w[o1, o2]')

    def tearDown(self):
        so.reset()




import sasoptpy
from sasoptpy.core import Variable, Constraint
from abc import ABC, abstractmethod

class Shadow(ABC):

    def __init__(self):
        self._abstract = True
        self._shadow = True

    @abstractmethod
    def _expr(self):
        pass


class ShadowVariable(Shadow, Variable):

    def __init__(self, name, **kwargs):
        Variable.__init__(self, name=name, internal=True)
        Shadow.__init__(self)

    def _initialize_self_coef(self):
        self.set_member(key=self._name + str(id(self)), ref=self, val=1)

    def set_group_key(self, vg, key):
        self._parent = vg
        self._iterkey = key
        if not sasoptpy.abstract.util.is_key_abstract(key):
            self._abstract = False

    def _expr(self):
        keylist = sasoptpy.util.package_utils._to_iterator_expression(
            self._iterkey)
        key = ', '.join(keylist)
        self_expr = '{}[{}]'.format(self._name, key)
        return self_expr

class ShadowConstraint(Shadow, Constraint):

    def __init__(self):
        super(Constraint, self).__init__()

    def _expr(self):
        return self._name
